"""Implementation of Valve learning layer 2/3 switch."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2018 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import random

from collections import defaultdict, deque

from faucet import tfm_pipeline
from faucet import valve_acl
from faucet import valve_flood
from faucet import valve_host
from faucet import valve_of
from faucet import valve_packet
from faucet import valve_route
from faucet import valve_util

from faucet.port import STACK_STATE_INIT, STACK_STATE_UP, STACK_STATE_DOWN
from faucet.vlan import NullVLAN


class ValveLogger(object):
    """Logger for a Valve that adds DP ID."""

    def __init__(self, logger, dp_id, dp_name):
        self.logger = logger
        self.dp_id = dp_id
        self.dp_name = dp_name

    def _dpid_prefix(self, log_msg):
        """Add DP ID prefix to log message."""
        return ' '.join((valve_util.dpid_log(self.dp_id), self.dp_name, log_msg))

    def debug(self, log_msg):
        """Log debug level message."""
        self.logger.debug(self._dpid_prefix(log_msg))

    def info(self, log_msg):
        """Log info level message."""
        self.logger.info(self._dpid_prefix(log_msg))

    def error(self, log_msg):
        """Log error level message."""
        self.logger.error(self._dpid_prefix(log_msg))

    def warning(self, log_msg):
        """Log warning level message."""
        self.logger.warning(self._dpid_prefix(log_msg))


class Valve(object):
    """Generates the messages to configure a datapath as a l2 learning switch.

    Vendor specific implementations may require sending configuration flows.
    This can be achieved by inheriting from this class and overwriting the
    function switch_features.
    """

    DEC_TTL = True
    USE_BARRIERS = True
    base_prom_labels = None
    recent_ofmsgs = deque(maxlen=32) # type: ignore
    logger = None
    ofchannel_logger = None
    host_manager = None
    flood_manager = None
    _route_manager_by_ipv = None
    _last_advertise_sec = None
    _port_highwater = {} # type: dict
    _last_update_metrics_sec = None
    _last_packet_in_sec = 0
    _packet_in_count_sec = 0

    def __init__(self, dp, logname, metrics, notifier):
        self.dp = dp
        self.logname = logname
        self.metrics = metrics
        self.notifier = notifier
        self.dp_init()

    def _skip_tables(self):
        tables = []
        any_routing = False
        for route_manager in list(self._route_manager_by_ipv.values()):
            if route_manager.active:
                any_routing = True
            else:
                tables.append(route_manager.fib_table)
        if not any_routing:
            tables.append(self.dp.tables['vip'])
        return tables

    def _active_tables(self):
        return set(self.dp.all_valve_tables()) - set(self._skip_tables())

    def close_logs(self):
        """Explicitly close any active loggers."""
        if self.logger is not None:
            valve_util.close_logger(self.logger.logger)
        valve_util.close_logger(self.ofchannel_logger)

    def dp_init(self):
        """Initialize datapath state at connection/re/config time."""
        self.close_logs()
        self.logger = ValveLogger(
            logging.getLogger(self.logname + '.valve'), self.dp.dp_id, self.dp.name)
        self.ofchannel_logger = None
        self.base_prom_labels = {
            'dp_id': hex(self.dp.dp_id),
            'dp_name': self.dp.name,
        }
        self._packet_in_count_sec = 0
        self._last_packet_in_sec = 0
        self._last_advertise_sec = 0
        self._route_manager_by_ipv = {}
        self._route_manager_by_eth_type = {}
        self._port_highwater = {}

        for vlan_vid in list(self.dp.vlans.keys()):
            self._port_highwater[vlan_vid] = {}
            for port_number in list(self.dp.ports.keys()):
                self._port_highwater[vlan_vid][port_number] = 0
        for fib_table, route_manager_class in (
                (self.dp.tables['ipv4_fib'], valve_route.ValveIPv4RouteManager),
                (self.dp.tables['ipv6_fib'], valve_route.ValveIPv6RouteManager)):
            route_manager = route_manager_class(
                self.logger, self.dp.arp_neighbor_timeout,
                self.dp.max_hosts_per_resolve_cycle, self.dp.max_host_fib_retry_count,
                self.dp.max_resolve_backoff_time, self.dp.proactive_learn, self.DEC_TTL,
                fib_table, self.dp.tables['vip'], self.dp.tables['eth_src'],
                self.dp.tables['eth_dst'], self.dp.tables['flood'],
                self.dp.highest_priority, self.dp.routers,
                self.dp.group_table_routing, self.dp.groups)
            self._route_manager_by_ipv[route_manager.IPV] = route_manager
            for vlan in list(self.dp.vlans.values()):
                if vlan.faucet_vips_by_ipv(route_manager.IPV):
                    route_manager.active = True
            for eth_type in route_manager.CONTROL_ETH_TYPES:
                self._route_manager_by_eth_type[eth_type] = route_manager
        if self.dp.stack:
            self.flood_manager = valve_flood.ValveFloodStackManager(
                self.dp.tables['flood'], self.dp.tables['eth_src'],
                self.dp.low_priority, self.dp.highest_priority,
                self.dp.group_table, self.dp.groups,
                self.dp.combinatorial_port_flood,
                self.dp.stack, self.dp.stack_ports,
                self.dp.shortest_path_to_root, self.dp.shortest_path_port)
        else:
            self.flood_manager = valve_flood.ValveFloodManager(
                self.dp.tables['flood'], self.dp.tables['eth_src'],
                self.dp.low_priority, self.dp.highest_priority,
                self.dp.group_table, self.dp.groups,
                self.dp.combinatorial_port_flood)
        if self.dp.use_idle_timeout:
            self.host_manager = valve_host.ValveHostFlowRemovedManager(
                self.logger, self.dp.ports, self.dp.vlans,
                self.dp.tables['eth_src'], self.dp.tables['eth_dst'],
                self.dp.timeout, self.dp.learn_jitter, self.dp.learn_ban_timeout,
                self.dp.low_priority, self.dp.highest_priority)
        else:
            self.host_manager = valve_host.ValveHostManager(
                self.logger, self.dp.ports, self.dp.vlans,
                self.dp.tables['eth_src'], self.dp.tables['eth_dst'],
                self.dp.timeout, self.dp.learn_jitter, self.dp.learn_ban_timeout,
                self.dp.low_priority, self.dp.highest_priority)

    def _notify(self, event_dict):
        """Send an event notification."""
        self.notifier.notify(self.dp.dp_id, self.dp.name, event_dict)

    def switch_features(self, _msg):
        """Send configuration flows necessary for the switch implementation.

        Args:
            msg (OFPSwitchFeatures): msg sent from switch.

        Vendor specific configuration should be implemented here.
        """
        return [
            valve_of.faucet_config(),
            valve_of.faucet_async(notify_flow_removed=self.dp.use_idle_timeout),
            valve_of.desc_stats_request()]

    def ofchannel_log(self, ofmsgs):
        """Log OpenFlow messages in text format to debugging log."""
        if self.dp is None:
            return
        if self.dp.ofchannel_log is None:
            return
        if self.ofchannel_logger is None:
            self.ofchannel_logger = valve_util.get_logger(
                self.dp.ofchannel_log,
                self.dp.ofchannel_log,
                logging.DEBUG,
                0)
        log_prefix = '%u %s' % (
            len(ofmsgs), valve_util.dpid_log(self.dp.dp_id))
        for i, ofmsg in enumerate(ofmsgs, start=1):
            self.ofchannel_logger.debug(
                '%u/%s %s', i, log_prefix, ofmsg)

    def _delete_all_valve_flows(self):
        """Delete all flows from all FAUCET tables."""
        ofmsgs = []
        ofmsgs.extend(self.dp.wildcard_table.flowdel())
        if self.dp.meters:
            ofmsgs.append(valve_of.meterdel())
        if self.dp.group_table:
            ofmsgs.append(self.dp.groups.delete_all())
        return ofmsgs

    def _delete_all_port_match_flows(self, port):
        """Delete all flows that match an input port from all FAUCET tables."""
        ofmsgs = []
        port_acl_table = self.dp.tables['port_acl']
        for table in self.dp.in_port_tables():
            if self.dp.dp_acls and table == port_acl_table:
                continue
            ofmsgs.extend(table.flowdel(
                match=table.match(in_port=port.number)))
        return ofmsgs

    def _add_default_drop_flows(self):
        """Add default drop rules on all FAUCET tables."""
        eth_src_table = self.dp.tables['eth_src']
        flood_table = self.dp.tables['flood']

        # default drop on all tables.
        ofmsgs = []
        for table in self._active_tables():
            ofmsgs.append(table.flowdrop(priority=self.dp.lowest_priority))

        # drop broadcast sources
        if self.dp.drop_broadcast_source_address:
            ofmsgs.append(eth_src_table.flowdrop(
                eth_src_table.match(eth_src=valve_of.mac.BROADCAST_STR),
                priority=self.dp.highest_priority))

        # antispoof for FAUCET's MAC address
        # TODO: antispoof for controller IPs on this VLAN, too.
        if self.dp.drop_spoofed_faucet_mac:
            for vlan in list(self.dp.vlans.values()):
                ofmsgs.append(eth_src_table.flowdrop(
                    eth_src_table.match(eth_src=vlan.faucet_mac),
                    priority=self.dp.high_priority))

        ofmsgs.append(flood_table.flowdrop(
            flood_table.match(
                eth_dst=valve_packet.CISCO_SPANNING_GROUP_ADDRESS),
            priority=self.dp.highest_priority))
        ofmsgs.append(flood_table.flowdrop(
            flood_table.match(
                eth_dst=valve_packet.BRIDGE_GROUP_ADDRESS,
                eth_dst_mask=valve_packet.BRIDGE_GROUP_MASK),
            priority=self.dp.highest_priority))

        return ofmsgs

    def _vlan_add_acl(self, vlan):
        ofmsgs = []
        if vlan.acls_in:
            acl_table = self.dp.tables['vlan_acl']
            acl_allow_inst = valve_of.goto_table(self.dp.tables['eth_src'])
            acl_force_port_vlan_inst = valve_of.goto_table(self.dp.tables['eth_dst'])
            ofmsgs = valve_acl.build_acl_ofmsgs(
                vlan.acls_in, acl_table,
                acl_allow_inst, acl_force_port_vlan_inst,
                self.dp.highest_priority, self.dp.meters,
                vlan.acls_in[0].exact_match, vlan_vid=vlan.vid)
        return ofmsgs

    def _add_vlan_flood_flow(self):
        """Add a flow to flood packets for unknown destinations."""
        return [self.dp.tables['eth_dst'].flowmod(
            priority=self.dp.low_priority,
            inst=[valve_of.goto_table(self.dp.tables['flood'])])]

    def _add_packetin_meter(self):
        """Add rate limiting of packet in pps (not supported by many DPs)."""
        if self.dp.packetin_pps:
            return [
                valve_of.controller_pps_meterdel(),
                valve_of.controller_pps_meteradd(pps=self.dp.packetin_pps)]
        return []

    def _add_dp_acls(self):
        """Add dataplane ACLs, if any."""
        ofmsgs = []
        if self.dp.dp_acls:
            port_acl_table = self.dp.tables['port_acl']
            acl_allow_inst = valve_of.goto_table(self.dp.tables['vlan'])
            acl_force_port_vlan_inst = valve_of.goto_table(self.dp.tables['eth_dst'])
            ofmsgs.extend(valve_acl.build_acl_ofmsgs(
                self.dp.dp_acls, port_acl_table,
                acl_allow_inst, acl_force_port_vlan_inst,
                self.dp.highest_priority, self.dp.meters,
                False)) # TODO: exact match support for DP ACLs.
        return ofmsgs

    def _add_non_local_vlan_destination_flow(self):
        """Add flow to handle packets not destined to a local VLAN."""
        return [self.dp.tables['eth_src'].flowmod(
            priority=self.dp.lowest_priority,
            inst=[valve_of.goto_table(self.dp.tables['eth_dst'])])]

    def _add_default_flows(self):
        """Configure datapath with necessary default tables and rules."""
        ofmsgs = []
        ofmsgs.extend(self._delete_all_valve_flows())
        ofmsgs.extend(self._add_packetin_meter())
        if self.dp.meters:
            for meter in list(self.dp.meters.values()):
                ofmsgs.append(meter.entry_msg)
        ofmsgs.extend(self._add_default_drop_flows())
        ofmsgs.extend(self._add_vlan_flood_flow())
        ofmsgs.extend(self._add_dp_acls())
        ofmsgs.extend(self._add_non_local_vlan_destination_flow())
        return ofmsgs

    def _add_vlan(self, vlan):
        """Configure a VLAN."""
        ofmsgs = []
        self.logger.info('Configuring %s' % vlan)
        # add ACL rules
        ofmsgs.extend(self._vlan_add_acl(vlan))
        # add controller IPs if configured.
        for ipv in vlan.ipvs():
            route_manager = self._route_manager_by_ipv[ipv]
            ofmsgs.extend(self._add_faucet_vips(
                route_manager, vlan, vlan.faucet_vips_by_ipv(ipv)))
        # install eth_dst_table flood ofmsgs
        ofmsgs.extend(self.flood_manager.build_flood_rules(vlan))
        # add learn rule for this VLAN.
        eth_src_table = self.dp.tables['eth_src']
        ofmsgs.append(eth_src_table.flowcontroller(
            eth_src_table.match(vlan=vlan),
            priority=self.dp.low_priority,
            inst=[valve_of.goto_table(self.dp.tables['eth_dst'])]))
        return ofmsgs

    def _del_vlan(self, vlan):
        """Delete a configured VLAN."""
        ofmsgs = []
        for table in self.dp.vlan_match_tables():
            ofmsgs.extend(table.flowdel(match=table.match(vlan=vlan)))
        self.logger.info('Delete VLAN %s' % vlan)
        return ofmsgs

    def _add_ports_and_vlans(self, discovered_up_port_nos):
        """Add all configured and discovered ports and VLANs."""
        all_configured_port_nos = set()

        for port in self.dp.stack_ports:
            all_configured_port_nos.add(port.number)

        for port in self.dp.output_only_ports:
            all_configured_port_nos.add(port.number)

        ofmsgs = []
        for vlan in list(self.dp.vlans.values()):
            vlan_ports = vlan.get_ports()
            if vlan_ports:
                for port in vlan_ports:
                    all_configured_port_nos.add(port.number)
                ofmsgs.extend(self._add_vlan(vlan))
            vlan.reset_caches()

        ports_status = defaultdict(bool)
        for port_no in discovered_up_port_nos:
            if port_no in all_configured_port_nos:
                ports_status[port_no] = True
        self._notify({'PORTS_STATUS': ports_status})

        all_up_port_nos = set()
        for port_no in all_configured_port_nos:
            if ports_status[port_no]:
                self._set_port_status(port_no, True)
                all_up_port_nos.add(port_no)
            else:
                self._set_port_status(port_no, False)

        ofmsgs.extend(
            self.ports_add(
                all_up_port_nos, cold_start=True, log_msg='configured'))
        self.dp.dyn_up_ports = set(discovered_up_port_nos)
        return ofmsgs

    def ofdescstats_handler(self, body):
        """Handle OF DP description."""
        self.metrics.of_dp_desc_stats.labels( # pylint: disable=no-member
            **dict(self.base_prom_labels,
                   mfr_desc=valve_util.utf8_decode(body.mfr_desc),
                   hw_desc=valve_util.utf8_decode(body.hw_desc),
                   sw_desc=valve_util.utf8_decode(body.sw_desc),
                   serial_num=valve_util.utf8_decode(body.serial_num),
                   dp_desc=valve_util.utf8_decode(body.dp_desc))).set(self.dp.dp_id)

    def _set_port_status(self, port_no, port_status):
        """Set port operational status."""
        port_labels = dict(self.base_prom_labels, port=port_no)
        self.metrics.port_status.labels( # pylint: disable=no-member
            **port_labels).set(port_status)
        if port_status:
            self.dp.dyn_up_ports.add(port_no)
        else:
            self.dp.dyn_up_ports -= set([port_no])

    def port_status_handler(self, port_no, reason, state):
        """Return OpenFlow messages responding to port operational status change."""

        def _decode_port_status(reason):
            """Humanize the port status reason code."""
            port_status_codes = {
                valve_of.ofp.OFPPR_ADD: 'ADD',
                valve_of.ofp.OFPPR_DELETE: 'DELETE',
                valve_of.ofp.OFPPR_MODIFY: 'MODIFY'
            }
            return port_status_codes.get(reason, 'UNKNOWN')

        port_status = valve_of.port_status_from_state(state)
        self._notify(
            {'PORT_CHANGE': {
                'port_no': port_no,
                'reason': _decode_port_status(reason),
                'state': state,
                'status': port_status}})
        ofmsgs = []
        self._set_port_status(port_no, port_status)
        if not self.port_no_valid(port_no):
            return ofmsgs
        port = self.dp.ports[port_no]
        if not port.opstatus_reconf:
            return ofmsgs
        if reason == valve_of.ofp.OFPPR_ADD:
            ofmsgs = self.port_add(port_no)
        elif reason == valve_of.ofp.OFPPR_DELETE:
            ofmsgs = self.port_delete(port_no)
        elif reason == valve_of.ofp.OFPPR_MODIFY:
            ofmsgs.extend(self.port_delete(port_no))
            if port_status:
                ofmsgs.extend(self.port_add(port_no))
        else:
            self.logger.warning('Unhandled port status %s/state %s for %s' % (
                reason, state, port))
        return ofmsgs

    def advertise(self, now):
        """Called periodically to advertise services (eg. IPv6 RAs)."""
        ofmsgs = []
        if (self.dp.advertise_interval and
                now - self._last_advertise_sec > self.dp.advertise_interval):
            for vlan in list(self.dp.vlans.values()):
                for route_manager in list(self._route_manager_by_ipv.values()):
                    ofmsgs.extend(route_manager.advertise(vlan))
            self._last_advertise_sec = now
        return ofmsgs

    def _send_lldp_beacon_on_port(self, port, now):
        chassis_id = str(self.dp.faucet_dp_mac)
        ttl = self.dp.lldp_beacon['send_interval'] * 3
        lldp_beacon = port.lldp_beacon
        chassis_id = str(self.dp.faucet_dp_mac)
        org_tlvs = [
            (tlv['oui'], tlv['subtype'], tlv['info'])
            for tlv in lldp_beacon['org_tlvs']]
        org_tlvs.extend(valve_packet.faucet_lldp_tlvs(self.dp))
        org_tlvs.extend(valve_packet.faucet_lldp_stack_state_tlvs(self.dp, port))
        system_name = lldp_beacon['system_name']
        if not system_name:
            system_name = self.dp.lldp_beacon['system_name']
        lldp_beacon_pkt = valve_packet.lldp_beacon(
            self.dp.faucet_dp_mac,
            chassis_id, port.number, ttl,
            org_tlvs=org_tlvs,
            system_name=system_name,
            port_descr=lldp_beacon['port_descr'])
        port.dyn_last_lldp_beacon_time = now
        return valve_of.packetout(port.number, lldp_beacon_pkt.data)

    def _lldp_beacon_ports(self, now):
        """Return list of ports to send LLDP packets; stacked ports always send LLDP."""
        priority_ports = set([port for port in self.dp.stack_ports if port.running() and port.lldp_beacon_enabled()])
        cutoff_beacon_time = now - self.dp.lldp_beacon['send_interval']
        nonpriority_ports = set([
            port for port in self.dp.lldp_beacon_ports
            if port.running() and (
                port.dyn_last_lldp_beacon_time is None or
                port.dyn_last_lldp_beacon_time < cutoff_beacon_time)])
        nonpriority_ports -= priority_ports
        send_ports = list(priority_ports)
        send_ports.extend(list(nonpriority_ports)[:self.dp.lldp_beacon['max_per_interval']])
        random.shuffle(send_ports)
        return send_ports

    def send_lldp_beacons(self, now):
        """Called periodically to send LLDP beacon packets."""
        # TODO: the beacon service is specifically NOT to support conventional R/STP.
        # It is intended to facilitate physical troubleshooting (e.g.
        # a standard cable tester can display OF port information).
        # It is used also by stacking to verify stacking links.
        # TODO: in the stacking case, provide an authentication scheme for the probes
        # so they cannot be forged.
        if not self.dp.lldp_beacon:
            return []
        send_ports = self._lldp_beacon_ports(now)
        self.logger.debug('sending LLDP beacons on ports %s' % send_ports)
        ofmsgs = [self._send_lldp_beacon_on_port(port, now) for port in send_ports]
        return ofmsgs

    def _update_stack_link_state(self, port, now):
        if port.is_stack_admin_down():
            return
        stack_probe_info = port.dyn_stack_probe_info
        last_seen_lldp_time = stack_probe_info.get('last_seen_lldp_time', None)
        if last_seen_lldp_time is None:
            return
        next_state = None
        remote_dp = port.stack['dp']
        stack_correct = stack_probe_info['stack_correct']
        remote_port_state = stack_probe_info['remote_port_state']
        send_interval = remote_dp.lldp_beacon['send_interval']
        num_lost_lldp = round((now - last_seen_lldp_time) / send_interval)
        if not stack_correct:
            if not port.is_stack_down():
                next_state = port.stack_down
                self.logger.error('Stack %s DOWN, incorrect cabling' % port)
        elif num_lost_lldp > port.max_lldp_lost:
            if not port.is_stack_down():
                next_state = port.stack_down
                self.logger.error(
                    'Stack %s DOWN, too many (%u) packets lost' % (port, num_lost_lldp))
        elif port.is_stack_down() and not port.is_stack_init():
            next_state = port.stack_init
            self.logger.info('Stack %s INIT' % port)
        elif (port.is_stack_init() and
              remote_port_state in set([STACK_STATE_UP, STACK_STATE_INIT])):
            next_state = port.stack_up
            self.logger.info('Stack %s UP' % port)
        elif port.is_stack_up() and remote_port_state == STACK_STATE_DOWN:
            next_state = port.stack_down
            self.logger.error('Stack %s DOWN, remote port is down' % port)
        if next_state is None:
            return
        next_state()
        port_labels = dict(self.base_prom_labels, port=port.number)
        self.metrics.port_stack_state.labels( # pylint: disable=no-member
            **port_labels).set(port.dyn_stack_current_state)
        port_stack_up = port.is_stack_up()
        self.flood_manager.update_stack_topo(port_stack_up, self.dp, port)

    def update_stack_link_states(self, now):
        """Called periodically to verify the state of stack ports."""
        for port in self.dp.stack_ports:
            self._update_stack_link_state(port, now)

    def datapath_connect(self, now, discovered_up_ports):
        """Handle Ryu datapath connection event and provision pipeline.

        Args:
            now (float): current epoch time.
            discovered_up_ports (list): datapath port numbers that are up.
        Returns:
            list: OpenFlow messages to send to datapath.
        """
        self.logger.info('Cold start configuring DP')
        self._notify(
            {'DP_CHANGE': {
                'reason': 'cold_start'}})
        ofmsgs = []
        ofmsgs.extend(self._add_default_flows())
        ofmsgs.extend(self._add_ports_and_vlans(discovered_up_ports))
        self.dp.dyn_last_coldstart_time = now
        self.dp.running = True
        self.metrics.of_dp_connections.labels( # pylint: disable=no-member
            **self.base_prom_labels).inc()
        self.metrics.dp_status.labels( # pylint: disable=no-member
            **self.base_prom_labels).set(1)
        return ofmsgs

    def datapath_disconnect(self):
        """Handle Ryu datapath disconnection event."""
        self.logger.warning('datapath down')
        self._notify(
            {'DP_CHANGE': {
                'reason': 'disconnect'}})
        self.dp.running = False
        self.metrics.of_dp_disconnections.labels( # pylint: disable=no-member
            **self.base_prom_labels).inc()
        self.metrics.dp_status.labels( # pylint: disable=no-member
            **self.base_prom_labels).set(0)

    def _port_add_acl(self, port, cold_start=False):
        ofmsgs = []
        acl_table = self.dp.tables['port_acl']
        in_port_match = acl_table.match(in_port=port.number)
        if cold_start:
            ofmsgs.extend(acl_table.flowdel(in_port_match))
        acl_allow_inst = valve_of.goto_table(self.dp.tables['vlan'])
        acl_force_port_vlan_inst = valve_of.goto_table(self.dp.tables['eth_dst'])
        if port.acls_in:
            ofmsgs.extend(valve_acl.build_acl_ofmsgs(
                port.acls_in, acl_table,
                acl_allow_inst, acl_force_port_vlan_inst,
                self.dp.highest_priority, self.dp.meters,
                port.acls_in[0].exact_match, port_num=port.number))
        else:
            ofmsgs.append(acl_table.flowmod(
                in_port_match,
                priority=self.dp.highest_priority,
                inst=[acl_allow_inst]))
        return ofmsgs

    def _port_add_vlan_rules(self, port, vlan_vid, vlan_inst):
        vlan_table = self.dp.tables['vlan']
        ofmsgs = []
        ofmsgs.append(vlan_table.flowmod(
            vlan_table.match(in_port=port.number, vlan=vlan_vid),
            priority=self.dp.low_priority,
            inst=vlan_inst))
        return ofmsgs

    def _port_add_vlan_untagged(self, port, vlan, forwarding_table, mirror_act):
        push_vlan_act = mirror_act + valve_of.push_vlan_act(vlan.vid)
        push_vlan_inst = [
            valve_of.apply_actions(push_vlan_act),
            valve_of.goto_table(forwarding_table)
        ]
        return self._port_add_vlan_rules(port, NullVLAN(), push_vlan_inst)

    def _port_add_vlan_tagged(self, port, vlan, forwarding_table, mirror_act):
        vlan_inst = [
            valve_of.goto_table(forwarding_table)
        ]
        if mirror_act:
            vlan_inst = [valve_of.apply_actions(mirror_act)] + vlan_inst
        return self._port_add_vlan_rules(port, vlan, vlan_inst)

    def _find_forwarding_table(self, vlan):
        if vlan.acls_in:
            return self.dp.tables['vlan_acl']
        return self.dp.tables['eth_src']

    def _port_add_vlans(self, port, mirror_act):
        ofmsgs = []
        for vlan in port.tagged_vlans:
            ofmsgs.extend(self._port_add_vlan_tagged(
                port, vlan, self._find_forwarding_table(vlan), mirror_act))
        if port.native_vlan is not None:
            ofmsgs.extend(self._port_add_vlan_untagged(
                port, port.native_vlan, self._find_forwarding_table(port.native_vlan), mirror_act))
        return ofmsgs

    def _port_delete_flows_state(self, port):
        """Delete flows/state for a port."""
        ofmsgs = []
        ofmsgs.extend(self._delete_all_port_match_flows(port))
        ofmsgs.extend(self.dp.tables['eth_dst'].flowdel(out_port=port.number))
        if port.permanent_learn:
            eth_src_table = self.dp.tables['eth_src']
            for entry in port.hosts():
                ofmsgs.extend(eth_src_table.flowdel(
                    match=eth_src_table.match(eth_src=entry.eth_src)))
        for vlan in port.vlans():
            vlan.clear_cache_hosts_on_port(port)
        return ofmsgs

    def ports_add(self, port_nums, cold_start=False, log_msg='up'):
        """Handle the addition of ports.

        Args:
            port_num (list): list of port numbers.
            cold_start (bool): True if configuring datapath from scratch.
        Returns:
            list: OpenFlow messages, if any.
        """
        ofmsgs = []
        vlans_with_ports_added = set()
        eth_src_table = self.dp.tables['eth_src']
        vlan_table = self.dp.tables['vlan']

        for port_num in port_nums:
            if port_num not in self.dp.ports:
                self.logger.info(
                    'Ignoring port:%u not present in configuration file' % port_num)
                continue

            port = self.dp.ports[port_num]
            port.dyn_phys_up = True
            self.logger.info('%s %s' % (port, log_msg))

            if not port.running():
                continue

            if port.output_only:
                ofmsgs.append(vlan_table.flowdrop(
                    match=vlan_table.match(in_port=port_num),
                    priority=self.dp.highest_priority))
                continue

            if port.receive_lldp:
                ofmsgs.append(vlan_table.flowcontroller(
                    match=vlan_table.match(
                        in_port=port_num,
                        eth_dst=valve_packet.LLDP_MAC_NEAREST_BRIDGE,
                        eth_dst_mask=valve_packet.BRIDGE_GROUP_MASK,
                        eth_type=valve_of.ether.ETH_TYPE_LLDP),
                    priority=self.dp.highest_priority,
                    max_len=128))

            if port.lacp:
                ofmsgs.extend(self.lacp_down(port))

            if port.override_output_port:
                ofmsgs.append(self.dp.tables['eth_src'].flowmod(
                    match=self.dp.tables['eth_src'].match(
                        in_port=port_num),
                    priority=self.dp.low_priority + 1,
                    inst=[valve_of.apply_actions([
                        valve_of.output_controller(),
                        valve_of.output_port(port.override_output_port.number)])]))

            if not self.dp.dp_acls:
                acl_ofmsgs = self._port_add_acl(port)
                ofmsgs.extend(acl_ofmsgs)

            port_vlans = port.vlans()

            # If this is a stacking port, accept all VLANs (came from another FAUCET)
            if port.stack is not None:
                # Actual stack traffic will have VLAN tags.
                ofmsgs.append(vlan_table.flowdrop(
                    match=vlan_table.match(
                        in_port=port_num,
                        vlan=NullVLAN()),
                    priority=self.dp.low_priority+1))
                ofmsgs.append(vlan_table.flowmod(
                    match=vlan_table.match(in_port=port_num),
                    priority=self.dp.low_priority,
                    inst=[valve_of.goto_table(eth_src_table)]))
                port_vlans = list(self.dp.vlans.values())
            else:
                mirror_act = port.mirror_actions()
                # Add port/to VLAN rules.
                ofmsgs.extend(self._port_add_vlans(port, mirror_act))

            for vlan in port_vlans:
                vlans_with_ports_added.add(vlan)

        # Only update flooding rules if not cold starting.
        if not cold_start:
            for vlan in vlans_with_ports_added:
                ofmsgs.extend(self.flood_manager.build_flood_rules(vlan))
        return ofmsgs

    def port_add(self, port_num):
        """Handle addition of a single port.

        Args:
            port_num (list): list of port numbers.
        Returns:
            list: OpenFlow messages, if any.
        """
        return self.ports_add([port_num])

    def ports_delete(self, port_nums, log_msg='down'):
        """Handle the deletion of ports.

        Args:
            port_nums (list): list of port numbers.
        Returns:
            list: OpenFlow messages, if any.
        """
        ofmsgs = []
        vlans_with_deleted_ports = set()

        for port_num in port_nums:
            if port_num not in self.dp.ports:
                continue
            port = self.dp.ports[port_num]
            port.dyn_phys_up = False
            self.logger.info('%s %s' % (port, log_msg))

            if not port.output_only:
                if port.lacp:
                    ofmsgs.extend(self.lacp_down(port))
                else:
                    ofmsgs.extend(self._port_delete_flows_state(port))
                for vlan in port.vlans():
                    vlans_with_deleted_ports.add(vlan)

        for vlan in vlans_with_deleted_ports:
            ofmsgs.extend(self.flood_manager.build_flood_rules(
                vlan, modify=True))

        return ofmsgs

    def port_delete(self, port_num):
        """Return flow messages that delete port from pipeline."""
        return self.ports_delete([port_num])

    def lacp_down(self, port):
        """Return OpenFlow messages when LACP is down on a port."""
        port.dyn_lacp_up = 0
        vlan_table = self.dp.tables['vlan']
        ofmsgs = []
        ofmsgs.extend(self._port_delete_flows_state(port))
        ofmsgs.append(vlan_table.flowdrop(
            match=vlan_table.match(in_port=port.number),
            priority=self.dp.high_priority))
        ofmsgs.append(vlan_table.flowcontroller(
            vlan_table.match(
                in_port=port.number,
                eth_type=valve_of.ether.ETH_TYPE_SLOW,
                eth_dst=valve_packet.SLOW_PROTOCOL_MULTICAST),
            priority=self.dp.highest_priority,
            max_len=128))
        self.metrics.port_lacp_status.labels( # pylint: disable=no-member
            **dict(self.base_prom_labels, port=port.number)).set(0)
        return ofmsgs

    def lacp_up(self, port):
        """Return OpenFlow messages when LACP is up on a port."""
        vlan_table = self.dp.tables['vlan']
        ofmsgs = []
        ofmsgs.extend(vlan_table.flowdel(
            match=vlan_table.match(in_port=port.number),
            priority=self.dp.high_priority, strict=True))
        self.metrics.port_lacp_status.labels( # pylint: disable=no-member
            **dict(self.base_prom_labels, port=port.number)).set(1)
        return ofmsgs

    def lacp_handler(self, now, pkt_meta):
        """Handle a LACP packet.

        We are a currently a passive, non-aggregateable LACP partner.

        Args:
            now (float): current epoch time.
            pkt_meta (PacketMeta): packet for control plane.
        Returns:
            list: OpenFlow messages, if any.
        """
        # TODO: ensure config consistent between LAG ports.
        ofmsgs = []
        if (pkt_meta.eth_dst == valve_packet.SLOW_PROTOCOL_MULTICAST and
                pkt_meta.eth_type == valve_of.ether.ETH_TYPE_SLOW and
                pkt_meta.port.lacp):
            pkt_meta.reparse_all()
            lacp_pkt = valve_packet.parse_lacp_pkt(pkt_meta.pkt)
            if lacp_pkt:
                last_lacp_up = pkt_meta.port.dyn_lacp_up
                pkt_meta.port.dyn_last_lacp_pkt = lacp_pkt
                pkt_meta.port.dyn_lacp_up = lacp_pkt.actor_state_synchronization
                pkt_meta.port.dyn_lacp_updated_time = now
                if last_lacp_up != pkt_meta.port.dyn_lacp_up:
                    self.logger.info('LACP state change from %s to %s on %s to %s LAG %u' % (
                        last_lacp_up, pkt_meta.port.dyn_lacp_up, pkt_meta.port,
                        lacp_pkt.actor_system, pkt_meta.port.lacp))
                    if pkt_meta.port.dyn_lacp_up:
                        ofmsgs.extend(self.lacp_up(pkt_meta.port))
                pkt = valve_packet.lacp_reqreply(
                    self.dp.faucet_dp_mac,
                    self.dp.faucet_dp_mac, pkt_meta.port.lacp, pkt_meta.port.number,
                    lacp_pkt.actor_system, lacp_pkt.actor_key, lacp_pkt.actor_port,
                    lacp_pkt.actor_system_priority, lacp_pkt.actor_port_priority,
                    lacp_pkt.actor_state_defaulted,
                    lacp_pkt.actor_state_expired,
                    lacp_pkt.actor_state_timeout,
                    lacp_pkt.actor_state_collecting,
                    lacp_pkt.actor_state_distributing,
                    lacp_pkt.actor_state_aggregation,
                    lacp_pkt.actor_state_synchronization,
                    lacp_pkt.actor_state_activity)
                ofmsgs.append(valve_of.packetout(pkt_meta.port.number, pkt.data))
        return ofmsgs

    @staticmethod
    def _get_tlvs_by_type(lldp_pkt, tlv_type):
        return [tlv for tlv in lldp_pkt.tlvs if tlv.tlv_type == tlv_type]

    @staticmethod
    def _tlvs_by_subtype(tlvs, subtype):
        return [tlv for tlv in tlvs if tlv.subtype == subtype]

    def _get_faucet_tlvs(self, lldp_pkt):
        return [tlv for tlv in self._get_tlvs_by_type(
            lldp_pkt, valve_packet.lldp.LLDP_TLV_ORGANIZATIONALLY_SPECIFIC)
                if tlv.oui == valve_packet.faucet_oui(self.dp.faucet_dp_mac)]

    def _parse_faucet_lldp(self, lldp_pkt):
        remote_dp_id = None
        remote_dp_name = None
        remote_port_id = None
        remote_port_state = None

        faucet_tlvs = self._get_faucet_tlvs(lldp_pkt)
        if faucet_tlvs:
            dp_id_tlvs = self._tlvs_by_subtype(
                faucet_tlvs, valve_packet.LLDP_FAUCET_DP_ID)
            dp_name_tlvs = self._get_tlvs_by_type(
                lldp_pkt, valve_packet.lldp.LLDP_TLV_SYSTEM_NAME)
            port_id_tlvs = self._get_tlvs_by_type(
                lldp_pkt, valve_packet.lldp.LLDP_TLV_PORT_ID)
            port_state_tlvs = self._tlvs_by_subtype(
                faucet_tlvs, valve_packet.LLDP_FAUCET_STACK_STATE)
            try:
                remote_dp_id = int(dp_id_tlvs[0].info)
                remote_port_id = int(port_id_tlvs[0].port_id)
                remote_port_state = int(port_state_tlvs[0].info)
                remote_dp_name = valve_util.utf8_decode(
                    dp_name_tlvs[0].system_name)
            except ValueError:
                pass
        return (remote_dp_id, remote_dp_name, remote_port_id, remote_port_state)

    def lldp_handler(self, now, pkt_meta):
        """Handle an LLDP packet.

        Args:
            pkt_meta (PacketMeta): packet for control plane.
        """
        if pkt_meta.eth_type != valve_of.ether.ETH_TYPE_LLDP:
            return
        pkt_meta.reparse_all()
        lldp_pkt = valve_packet.parse_lldp(pkt_meta.pkt)
        if not lldp_pkt:
            return

        port = pkt_meta.port
        (remote_dp_id, remote_dp_name,
         remote_port_id, remote_port_state) = self._parse_faucet_lldp(lldp_pkt)

        if remote_dp_id and remote_port_id:
            self.logger.info('FAUCET LLDP from %s (remote %s, port %u)' % (
                pkt_meta.port, valve_util.dpid_log(remote_dp_id), remote_port_id))
            if port.stack:
                remote_dp = port.stack['dp']
                remote_port = port.stack['port']
                stack_correct = True
                self.metrics.stack_probes_received.labels( # pylint: disable=no-member
                    **self.base_prom_labels).inc()
                if (remote_dp_id != remote_dp.dp_id or
                        remote_dp_name != remote_dp.name or
                        remote_port_id != remote_port.number):
                    self.logger.error(
                        'Stack %s cabling incorrect, expected %s:%s:%u, actual %s:%s:%u' % (
                            port,
                            valve_util.dpid_log(remote_dp.dp_id),
                            remote_dp.name,
                            remote_port.number,
                            valve_util.dpid_log(remote_dp_id),
                            remote_dp_name,
                            remote_port_id))
                    stack_correct = False
                    self.metrics.stack_cabling_errors.labels( # pylint: disable=no-member
                        **self.base_prom_labels).inc()
                port.dyn_stack_probe_info = {
                    'last_seen_lldp_time': now,
                    'stack_correct': stack_correct,
                    'remote_dp_id': remote_dp_id,
                    'remote_dp_name': remote_dp_name,
                    'remote_port_id': remote_port_id,
                    'remote_port_state': remote_port_state
                }
                self._update_stack_link_state(port, now)

        self.logger.debug('LLDP from %s: %s' % (pkt_meta.port, str(lldp_pkt)))

    @staticmethod
    def _control_plane_handler(now, pkt_meta, route_manager):
        """Handle a packet probably destined to FAUCET's route managers.

        For example, next hop resolution or ICMP echo requests.

        Args:
            pkt_meta (PacketMeta): packet for control plane.
            route_manager (ValveRouteManager): route manager for this eth_type.
        Returns:
            list: OpenFlow messages, if any.
        """
        if (pkt_meta.eth_dst == pkt_meta.vlan.faucet_mac or
                not valve_packet.mac_addr_is_unicast(pkt_meta.eth_dst)):
            return route_manager.control_plane_handler(now, pkt_meta)
        return []

    def rate_limit_packet_ins(self, now):
        """Return True if too many packet ins this second."""
        if self._last_packet_in_sec != now:
            self._last_packet_in_sec = now
            self._packet_in_count_sec = 0
        self._packet_in_count_sec += 1
        if self.dp.ignore_learn_ins:
            if self._packet_in_count_sec % self.dp.ignore_learn_ins == 0:
                self.metrics.of_ignored_packet_ins.labels( # pylint: disable=no-member
                    **self.base_prom_labels).inc()
                return True
        return False

    def _learn_host(self, now, other_valves, pkt_meta):
        """Possibly learn a host on a port.

        Args:
            valves (list): of all Valves (datapaths).
            pkt_meta (PacketMeta): PacketMeta instance for packet received.
        Returns:
            list: OpenFlow messages, if any.
        """
        learn_port = self.flood_manager.edge_learn_port(
            other_valves, pkt_meta)
        if learn_port is not None:
            learn_flows, previous_port = self.host_manager.learn_host_on_vlan_ports(
                now, learn_port, pkt_meta.vlan, pkt_meta.eth_src,
                last_dp_coldstart_time=self.dp.dyn_last_coldstart_time)
            if learn_flows:
                if pkt_meta.l3_pkt is None:
                    pkt_meta.reparse_ip()
                previous_port_no = None
                port_move_text = ''
                if previous_port is not None:
                    previous_port_no = previous_port.number
                    if pkt_meta.port.number != previous_port_no:
                        port_move_text = ', moved from port %u' % previous_port_no
                self.logger.info(
                    'L2 learned %s (L2 type 0x%4.4x, L3 src %s, L3 dst %s) '
                    'on %s%s on VLAN %u (%u hosts total)' % (
                        pkt_meta.eth_src, pkt_meta.eth_type,
                        pkt_meta.l3_src, pkt_meta.l3_dst, pkt_meta.port, port_move_text,
                        pkt_meta.vlan.vid, pkt_meta.vlan.hosts_count()))
                self._notify(
                    {'L2_LEARN': {
                        'port_no': pkt_meta.port.number,
                        'previous_port_no': previous_port_no,
                        'vid': pkt_meta.vlan.vid,
                        'eth_src': pkt_meta.eth_src,
                        'eth_type': pkt_meta.eth_type,
                        'l3_src_ip': pkt_meta.l3_src,
                        'l3_dst_ip': pkt_meta.l3_dst}})
                return learn_flows
        return []

    def port_no_valid(self, port_no):
        """Return True if supplied port number valid on this datapath."""
        if valve_of.ignore_port(port_no):
            return False
        if port_no not in self.dp.ports:
            self.logger.info('port %u unknown' % port_no)
            return False
        return True

    def parse_rcv_packet(self, in_port, vlan_vid, eth_type, data, orig_len, pkt, eth_pkt):
        """Parse a received packet into a PacketMeta instance.

        Args:
            in_port (int): port packet was received on.
            vlan_vid (int): VLAN VID of port packet was received on.
            eth_type (int): Ethernet type of packet.
            data (bytes): Raw packet data.
            orig_len (int): Original length of packet.
            pkt (ryu.lib.packet.packet): parsed packet received.
            ekt_pkt (ryu.lib.packet.ethernet): parsed Ethernet header.
        Returns:
            PacketMeta instance.
        """
        eth_src = eth_pkt.src
        eth_dst = eth_pkt.dst
        vlan = None
        if vlan_vid is not None:
            vlan = self.dp.vlans[vlan_vid]
        port = self.dp.ports[in_port]
        return valve_packet.PacketMeta(
            data, orig_len, pkt, eth_pkt, port, vlan, eth_src, eth_dst, eth_type)

    def parse_pkt_meta(self, msg):
        """Parse OF packet-in message to PacketMeta."""
        if not self.dp.running:
            return None
        if self.dp.cookie != msg.cookie:
            return None
        # Drop any packet we didn't specifically ask for
        if msg.reason != valve_of.ofp.OFPR_ACTION:
            return None
        if not msg.match:
            return None
        in_port = msg.match['in_port']
        if not in_port or not self.port_no_valid(in_port):
            return None

        if not msg.data:
            return None
        # Truncate packet in data (OVS > 2.5 does not honor max_len)
        data = msg.data[:valve_of.MAX_PACKET_IN_BYTES]

        # eth/VLAN header only
        pkt, eth_pkt, eth_type, vlan_vid = valve_packet.parse_packet_in_pkt(
            data, max_len=valve_packet.ETH_VLAN_HEADER_SIZE)
        if pkt is None or eth_pkt is None:
            self.logger.info(
                'unparseable packet from port %u' % in_port)
            return None
        if vlan_vid is not None and vlan_vid not in self.dp.vlans:
            self.logger.info(
                'packet for unknown VLAN %u' % vlan_vid)
            return None
        pkt_meta = self.parse_rcv_packet(
            in_port, vlan_vid, eth_type, data, msg.total_len, pkt, eth_pkt)
        if not valve_packet.mac_addr_is_unicast(pkt_meta.eth_src):
            self.logger.info(
                'packet with non-unicast eth_src %s port %u' % (
                    pkt_meta.eth_src, in_port))
            return None
        if self.dp.stack is not None:
            if (not pkt_meta.port.stack and
                    pkt_meta.vlan and
                    pkt_meta.vlan not in pkt_meta.port.tagged_vlans and
                    pkt_meta.vlan != pkt_meta.port.native_vlan):
                self.logger.warning(
                    ('packet from non-stack port number %u is not member of VLAN %u' % (
                        pkt_meta.port.number, pkt_meta.vlan.vid)))
                return None
        return pkt_meta

    def update_config_metrics(self):
        """Update gauge/metrics for configuration."""
        self.metrics.reset_dpid(self.base_prom_labels)
        for table_id, table in list(self.dp.tables_by_id.items()):
            self.metrics.faucet_config_table_names.labels(
                **dict(self.base_prom_labels, table_name=table.name)).set(table_id)

    def update_metrics(self, now, updated_port=None, rate_limited=False):
        """Update Gauge/metrics."""
        if self._last_update_metrics_sec and rate_limited:
            if now - self._last_update_metrics_sec < self.dp.metrics_rate_limit_sec:
                return
        self._last_update_metrics_sec = now

        def _update_vlan(vlan):
            vlan_labels = dict(self.base_prom_labels, vlan=vlan.vid)
            self.metrics.vlan_hosts_learned.labels(
                **vlan_labels).set(vlan.hosts_count())
            self.metrics.vlan_learn_bans.labels(
                **vlan_labels).set(vlan.dyn_learn_ban_count)
            for ipv in vlan.ipvs():
                self.metrics.vlan_neighbors.labels(
                    **dict(vlan_labels, ipv=ipv)).set(vlan.neigh_cache_count_by_ipv(ipv))

        def _update_port(vlan, port):
            port_labels = dict(self.base_prom_labels, port=port.number)
            port_vlan_labels = dict(self.base_prom_labels, vlan=vlan.vid, port=port.number)
            port_vlan_hosts_learned = port.hosts_count(vlans=[vlan])
            self.metrics.port_vlan_hosts_learned.labels(
                **port_vlan_labels).set(port_vlan_hosts_learned)
            self.metrics.port_learn_bans.labels(
                **port_labels).set(port.dyn_learn_ban_count)
            highwater = self._port_highwater[vlan.vid][port.number]
            if highwater > port_vlan_hosts_learned:
                for i in range(port_vlan_hosts_learned, highwater + 1):
                    self.metrics.learned_macs.labels(
                        **dict(port_vlan_labels, n=i)).set(0)
            self._port_highwater[vlan.vid][port.number] = port_vlan_hosts_learned
            port_vlan_hosts = port.hosts(vlans=[vlan])
            assert port_vlan_hosts_learned == len(port_vlan_hosts)
            # TODO: make MAC table updates less expensive.
            for i, entry in enumerate(sorted(port_vlan_hosts)):
                self.metrics.learned_macs.labels(
                    **dict(port_vlan_labels, n=i)).set(entry.eth_src_int)

        if updated_port:
            for vlan in updated_port.vlans():
                _update_vlan(vlan)
                _update_port(vlan, updated_port)
        else:
            for vlan in list(self.dp.vlans.values()):
                _update_vlan(vlan)
                for port in vlan.get_ports():
                    _update_port(vlan, port)

    def rcv_packet(self, now, other_valves, pkt_meta):
        """Handle a packet from the dataplane (eg to re/learn a host).

        The packet may be sent to us also in response to FAUCET
        initiating IPv6 neighbor discovery, or ARP, to resolve
        a nexthop.

        Args:
            other_valves (list): all Valves other than this one.
            pkt_meta (PacketMeta): packet for control plane.
        Return:
            list: OpenFlow messages, if any.
        """
        ofmsgs = []

        self.logger.debug(
            'Packet_in src:%s in_port:%d VLAN:%s' % (
                pkt_meta.eth_src,
                pkt_meta.port.number,
                pkt_meta.vlan))

        if pkt_meta.vlan is None:
            self.metrics.of_non_vlan_packet_ins.labels( # pylint: disable=no-member
                **self.base_prom_labels).inc()
            if pkt_meta.port.lacp:
                lacp_ofmsgs = self.lacp_handler(now, pkt_meta)
                if lacp_ofmsgs:
                    return lacp_ofmsgs
            self.lldp_handler(now, pkt_meta)
            # TODO: verify LLDP message (e.g. org-specific authenticator TLV)
            return ofmsgs

        self.metrics.of_vlan_packet_ins.labels( # pylint: disable=no-member
            **self.base_prom_labels).inc()

        ban_rules = self.host_manager.ban_rules(pkt_meta)
        if ban_rules:
            return ban_rules

        if pkt_meta.eth_type in self._route_manager_by_eth_type:
            route_manager = self._route_manager_by_eth_type[pkt_meta.eth_type]
            if route_manager.active:
                pkt_meta.reparse_ip()
                if pkt_meta.l3_pkt:
                    control_plane_ofmsgs = self._control_plane_handler(now, pkt_meta, route_manager)
                    if control_plane_ofmsgs:
                        ofmsgs.extend(control_plane_ofmsgs)
                    else:
                        ofmsgs.extend(route_manager.add_host_fib_route_from_pkt(now, pkt_meta))

        ofmsgs.extend(self._learn_host(now, other_valves, pkt_meta))
        return ofmsgs

    def _lacp_state_expire(self, vlan, now):
        """Expire controller state for LACP.

        Args:
            vlan (VLAN instance): VLAN with LAGs.
            now (float): current epoch time.
        Return:
            list: OpenFlow messages, if any.
        """
        ofmsgs = []
        for ports in list(vlan.lags().values()):
            lacp_up_ports = [port for port in ports if port.dyn_lacp_up]
            for port in lacp_up_ports:
                lacp_age = now - port.dyn_lacp_updated_time
                if lacp_age > self.dp.lacp_timeout:
                    self.logger.info('LACP on %s expired' % port)
                    ofmsgs.extend(self.lacp_down(port))
        return ofmsgs

    def state_expire(self, now):
        """Expire controller caches/state (e.g. hosts learned).

        Expire state from the host manager only; the switch does its own flow
        expiry.

        Return:
            list: OpenFlow messages, if any.
        """
        ofmsgs = []
        if self.dp.running:
            for vlan in list(self.dp.vlans.values()):
                expired_hosts = self.host_manager.expire_hosts_from_vlan(vlan, now)
                for entry in expired_hosts:
                    self._notify(
                        {'L2_EXPIRE': {
                            'port_no': entry.port.number,
                            'vid': vlan.vid,
                            'eth_src': entry.eth_src}})
                self._lacp_state_expire(vlan, now)
        return ofmsgs

    def _apply_config_changes(self, new_dp, changes):
        """Apply any detected configuration changes.

        Args:
            new_dp: (DP): new dataplane configuration.
            changes (tuple) of:
                deleted_ports (list): deleted port numbers.
                changed_ports (list): changed/added port numbers.
                changed_acl_ports (set): changed ACL only port numbers.
                deleted_vlans (list): deleted VLAN IDs.
                changed_vlans (list): changed/added VLAN IDs.
                all_ports_changed (bool): True if all ports changed.
        Returns:
            tuple:
                cold_start (bool): whether cold starting.
                ofmsgs (list): OpenFlow messages.
        """
        new_dp.running = True
        (deleted_ports, changed_ports, changed_acl_ports,
         deleted_vlans, changed_vlans, all_ports_changed) = changes

        if all_ports_changed:
            self.logger.info('all ports changed')
            self.dp = new_dp
            self.dp_init()
            return True, []

        ofmsgs = []

        if deleted_ports:
            self.logger.info('ports deleted: %s' % deleted_ports)
            ofmsgs.extend(self.ports_delete(deleted_ports))
        if deleted_vlans:
            self.logger.info('VLANs deleted: %s' % deleted_vlans)
            for vid in deleted_vlans:
                vlan = self.dp.vlans[vid]
                ofmsgs.extend(self._del_vlan(vlan))
        if changed_ports:
            self.logger.info('ports changed/added: %s' % changed_ports)
            ofmsgs.extend(self.ports_delete(changed_ports))

        self.dp = new_dp
        self.dp.reset_refs()
        self.dp_init()

        if changed_vlans:
            self.logger.info('VLANs changed/added: %s' % changed_vlans)
            for vid in changed_vlans:
                vlan = self.dp.vlans[vid]
                ofmsgs.extend(self._del_vlan(vlan))
                ofmsgs.extend(self._add_vlan(vlan))
        if changed_ports:
            ofmsgs.extend(self.ports_add(changed_ports))
        if changed_acl_ports:
            self.logger.info('ports with ACL only changed: %s' % changed_acl_ports)
            for port_num in changed_acl_ports:
                port = self.dp.ports[port_num]
                ofmsgs.extend(self._port_add_acl(port, cold_start=True))

        return False, ofmsgs

    def reload_config(self, now, new_dp):
        """Reload configuration new_dp.

        Following config changes are currently supported:
            - Port config: support all available configs
                  (e.g. native_vlan, acl_in) & change operations
                  (add, delete, modify) a port
            - ACL config:support any modification, currently reload all
                  rules belonging to an ACL
            - VLAN config: enable, disable routing, etc...

        Args:
            now (float): current epoch time.
            new_dp (DP): new dataplane configuration.
        Returns:
            ofmsgs (list): OpenFlow messages.
        """
        dp_running = self.dp.running
        up_ports = self.dp.dyn_up_ports
        cold_start, ofmsgs = self._apply_config_changes(
            new_dp, self.dp.get_config_changes(self.logger, new_dp))
        self.dp.running = dp_running
        restart_type = 'none'
        if self.dp.running:
            if cold_start:
                # Need to reprovision pipeline on cold start.
                ofmsgs = self.switch_features(None) + self.datapath_connect(now, up_ports)
            if ofmsgs:
                if cold_start:
                    self.metrics.faucet_config_reload_cold.labels( # pylint: disable=no-member
                        **self.base_prom_labels).inc()
                    self.logger.info('Cold starting')
                    restart_type = 'cold'
                else:
                    self.metrics.faucet_config_reload_warm.labels( # pylint: disable=no-member
                        **self.base_prom_labels).inc()
                    self.logger.info('Warm starting')
                    restart_type = 'warm'
        else:
            ofmsgs = []
        self._notify({'CONFIG_CHANGE': {'restart_type': restart_type}})
        return ofmsgs

    @staticmethod
    def _add_faucet_vips(route_manager, vlan, faucet_vips):
        ofmsgs = []
        for faucet_vip in faucet_vips:
            ofmsgs.extend(route_manager.add_faucet_vip(vlan, faucet_vip))
        return ofmsgs

    def add_authed_mac(self, port_num, mac):
        """Add authed mac address"""
        # TODO: track dynamic auth state.
        ofmsg = self.dp.tables['port_acl'].flowmod(
            self.dp.tables['port_acl'].match(
                in_port=port_num,
                eth_src=mac),
            priority=self.dp.highest_priority,
            inst=[valve_of.goto_table(self.dp.tables['vlan'])])
        return [ofmsg]

    def add_route(self, vlan, ip_gw, ip_dst):
        """Add route to VLAN routing table."""
        route_manager = self._route_manager_by_ipv[ip_dst.version]
        return route_manager.add_route(vlan, ip_gw, ip_dst)

    def del_route(self, vlan, ip_dst):
        """Delete route from VLAN routing table."""
        route_manager = self._route_manager_by_ipv[ip_dst.version]
        return route_manager.del_route(vlan, ip_dst)

    def resolve_gateways(self, now):
        """Call route managers to re/resolve gateways.

        Returns:
            list: OpenFlow messages, if any.
        """
        if not self.dp.running:
            return []
        ofmsgs = []
        for vlan in list(self.dp.vlans.values()):
            for route_manager in list(self._route_manager_by_ipv.values()):
                ofmsgs.extend(route_manager.resolve_gateways(vlan, now))
        return ofmsgs

    def oferror(self, msg):
        """Correlate OFError message with flow we sent, if any.

        Args:
            msg (ryu.controller.ofp_event.EventOFPMsgBase): message from datapath.
        """
        self.metrics.of_errors.labels( # pylint: disable=no-member
            **self.base_prom_labels).inc()
        orig_msgs = [orig_msg for orig_msg in self.recent_ofmsgs if orig_msg.xid == msg.xid]
        error_txt = msg
        if orig_msgs:
            error_txt = orig_msgs[0]
        self.logger.error('OFError %s' % error_txt)

    def prepare_send_flows(self, flow_msgs):
        """Prepare to send flows to datapath.

        Args:
            flow_msgs (list): OpenFlow messages to send.
        """
        reordered_flow_msgs = valve_of.valve_flowreorder(
            flow_msgs, use_barriers=self.USE_BARRIERS)
        self.ofchannel_log(reordered_flow_msgs)
        self.metrics.of_flowmsgs_sent.labels( # pylint: disable=no-member
            **self.base_prom_labels).inc(len(reordered_flow_msgs))
        self.recent_ofmsgs.extend(reordered_flow_msgs)
        return reordered_flow_msgs

    def send_flows(self, ryu_dp, flow_msgs):
        """Send flows to datapath.

        Args:
            ryu_dp (ryu.controller.controller.Datapath): datapath.
            flow_msgs (list): OpenFlow messages to send.
        """
        for flow_msg in self.prepare_send_flows(flow_msgs):
            flow_msg.datapath = ryu_dp
            ryu_dp.send_msg(flow_msg)

    def flow_timeout(self, now, table_id, match):
        """Call flow timeout message handler:

        Args:
            now (float): current epoch time.
            table_id (int): ID of table where flow was installed.
            match (dict): match conditions for expired flow.
        Returns:
            list: OpenFlow messages, if any.
        """
        return self.host_manager.flow_timeout(now, table_id, match)

    def get_config_dict(self):
        """Return datapath config as a dict for experimental API."""
        return self.dp.get_config_dict()


class TfmValve(Valve):
    """Valve implementation that uses OpenFlow send table features messages."""

    PIPELINE_CONF = 'tfm_pipeline.json'
    SKIP_VALIDATION_TABLES = ()

    def _verify_pipeline_config(self, tfm):
        for tfm_table in tfm.body:
            table = self.dp.tables_by_id[tfm_table.table_id]
            if table.table_id in self.SKIP_VALIDATION_TABLES:
                continue
            if table.restricted_match_types is None:
                continue
            for prop in tfm_table.properties:
                if not (isinstance(prop, valve_of.parser.OFPTableFeaturePropOxm)
                        and prop.type == 8):
                    continue
                tfm_matches = set(sorted([oxm.type for oxm in prop.oxm_ids]))
                if tfm_matches != table.restricted_match_types:
                    self.logger.info(
                        'table %s ID %s match TFM config %s != pipeline %s' % (
                            tfm_table.name, tfm_table.table_id,
                            tfm_matches, table.restricted_match_types))

    def switch_features(self, msg):
        ofmsgs = self._delete_all_valve_flows()
        ofmsgs.extend(super(TfmValve, self).switch_features(msg))
        ryu_table_loader = tfm_pipeline.LoadRyuTables(
            self.dp.pipeline_config_dir, self.PIPELINE_CONF)
        self.logger.info('loading pipeline configuration')
        active_table_ids = [table.table_id for table in self._active_tables()]
        tfm = valve_of.table_features(
            ryu_table_loader.load_tables(active_table_ids))
        self._verify_pipeline_config(tfm)
        ofmsgs.append(tfm)
        return ofmsgs


class ArubaValve(TfmValve):
    """Valve implementation that uses OpenFlow send table features messages."""

    PIPELINE_CONF = 'aruba_pipeline.json'
    DEC_TTL = False

class CiscoC9KValve(TfmValve):
    """Valve implementation that uses OpenFlow send table features messages."""

    PIPELINE_CONF = 'cisco_c9k_pipeline.json'

class OVSValve(Valve):
    """Valve implementation for OVS."""

    USE_BARRIERS = False


class AlliedTelesis(OVSValve):
    """Valve implementation for AT."""

    DEC_TTL = False


SUPPORTED_HARDWARE = {
    'Allied-Telesis': AlliedTelesis,
    'Aruba': ArubaValve,
    'CiscoC9K': CiscoC9KValve,
    'GenericTFM': TfmValve,
    'Lagopus': Valve,
    'Netronome': Valve,
    'NoviFlow': Valve,
    'Open vSwitch': OVSValve,
    'ZodiacFX': Valve,
}


def valve_factory(dp):
    """Return a Valve object based dp's hardware configuration field.

    Args:
        dp (DP): DP instance with the configuration for this Valve.
    """
    if dp.hardware in SUPPORTED_HARDWARE:
        return SUPPORTED_HARDWARE[dp.hardware]
    return None



"""VLAN configuration."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2018 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import ipaddress
import random

import netaddr

from faucet import valve_of
from faucet.conf import Conf, test_config_condition, InvalidConfigError
from faucet.valve_util import btos
from faucet.valve_packet import FAUCET_MAC


class NullVLAN(object):
    """Placeholder null VLAN."""

    name = 'Null VLAN'
    vid = valve_of.ofp.OFPVID_NONE


class HostCacheEntry(object):
    """Association of a host with a port."""

    def __init__(self, eth_src, port, cache_time):
        self.eth_src = eth_src
        self.port = port
        self.cache_time = cache_time
        self.eth_src_int = int(eth_src.replace(':', ''), 16)

    def __hash__(self):
        return hash((self.eth_src_int, self.port.number))

    def __str__(self):
        return '%s on %s' % (self.eth_src, self.port)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __lt__(self, other):
        return self.__hash__() < other.__hash__()


class VLAN(Conf):
    """Contains state for one VLAN, including its configuration."""

# Note: while vlans are configured once for each datapath, there will be a
# separate vlan object created for each datapath that the vlan appears on

    name = None
    dp_id = None
    tagged = None
    untagged = None
    vid = None
    faucet_vips = None
    faucet_mac = None
    bgp_as = None
    bgp_connect_mode = None
    bgp_local_address = None
    bgp_server_addresses = [] # type: list
    bgp_port = None
    bgp_routerid = None
    bgp_neighbor_addresses = [] # type: list
    bgp_neighbour_addresses = [] # type: list
    bgp_neighbor_as = None
    bgp_neighbour_as = None
    routes = None
    max_hosts = None
    unicast_flood = None
    acl_in = None
    acls_in = None
    targeted_gw_resolution = None
    proactive_arp_limit = None
    proactive_nd_limit = None
    # Define dynamic variables with prefix dyn_ to distinguish from variables set
    # configuration
    dyn_host_cache = None
    dyn_host_cache_by_port = None
    dyn_faucet_vips_by_ipv = None
    dyn_bgp_server_addresses_by_ipv = None
    dyn_bgp_neighbor_addresses_by_ipv = None
    dyn_routes_by_ipv = None
    dyn_gws_by_ipv = None
    dyn_neigh_cache_by_ipv = None
    dyn_learn_ban_count = 0
    dyn_last_time_hosts_expired = None
    dyn_oldest_host_time = None

    defaults = {
        'name': None,
        'description': None,
        'acl_in': None,
        'acls_in': None,
        'faucet_vips': None,
        'faucet_mac': FAUCET_MAC,
        # set MAC for FAUCET VIPs on this VLAN
        'unicast_flood': True,
        'bgp_as': None,
        'bgp_connect_mode': 'both',
        'bgp_local_address': None,
        'bgp_port': 9179,
        'bgp_server_addresses': ['0.0.0.0', '::'],
        'bgp_routerid': None,
        'bgp_neighbour_addresses': [],
        'bgp_neighbor_addresses': [],
        'bgp_neighbour_as': None,
        'bgp_neighbor_as': None,
        'routes': None,
        'max_hosts': 256,
        # Limit number of hosts that can be learned on a VLAN.
        'vid': None,
        'proactive_arp_limit': 0,
        # Don't proactively ARP for hosts if over this limit (default 2*max_hosts)
        'proactive_nd_limit': 0,
        # Don't proactively ND for hosts if over this limit (default 2*max_hosts)
        'targeted_gw_resolution': False,
        # If True, and a gateway has been resolved, target the first re-resolution attempt to the same port rather than flooding.
        'minimum_ip_size_check': True,
        # If False, don't check that IP packets have a payload (must be False for OVS trace/tutorial to work)
        }

    defaults_types = {
        'name': str,
        'description': str,
        'acl_in': (int, str),
        'acls_in': list,
        'faucet_vips': list,
        'faucet_mac': str,
        'unicast_flood': bool,
        'bgp_as': int,
        'bgp_connect_mode': str,
        'bgp_local_address': str,
        'bgp_port': int,
        'bgp_server_addresses': list,
        'bgp_routerid': str,
        'bgp_neighbour_addresses': list,
        'bgp_neighbor_addresses': list,
        'bgp_neighbour_as': int,
        'bgp_neighbor_as': int,
        'routes': list,
        'max_hosts': int,
        'vid': int,
        'proactive_arp_limit': int,
        'proactive_nd_limit': int,
        'targeted_gw_resolution': bool,
        'minimum_ip_size_check': bool,
    }

    def __init__(self, _id, dp_id, conf=None):
        self.tagged = []
        self.untagged = []
        self.dyn_faucet_vips_by_ipv = collections.defaultdict(list)
        self.dyn_bgp_server_addresses_by_ipv = collections.defaultdict(list)
        self.dyn_bgp_neighbor_addresses_by_ipv = collections.defaultdict(list)
        self.dyn_routes_by_ipv = collections.defaultdict(dict)
        self.dyn_gws_by_ipv = collections.defaultdict(dict)
        self.dyn_ipvs = []
        self.dyn_bgp_ipvs = []
        self.reset_caches()
        super(VLAN, self).__init__(_id, dp_id, conf)

    def set_defaults(self):
        super(VLAN, self).set_defaults()
        self._set_default('vid', self._id)
        self._set_default('name', str(self._id))
        self._set_default('faucet_vips', [])
        self._set_default('bgp_neighbor_as', self.bgp_neighbour_as)
        self._set_default(
            'bgp_neighbor_addresses', self.bgp_neighbour_addresses)

    def check_config(self):
        super(VLAN, self).check_config()
        test_config_condition(not self.vid_valid(self.vid), 'invalid VID %s' % self.vid)
        test_config_condition(not netaddr.valid_mac(self.faucet_mac), (
            'invalid MAC address %s' % self.faucet_mac))
        if self.max_hosts:
            if not self.proactive_arp_limit:
                self.proactive_arp_limit = 2 * self.max_hosts
            if not self.proactive_nd_limit:
                self.proactive_nd_limit = 2 * self.max_hosts
        if self.faucet_vips:
            try:
                self.faucet_vips = [
                    ipaddress.ip_interface(btos(ip)) for ip in self.faucet_vips]
            except (ValueError, AttributeError, TypeError) as err:
                raise InvalidConfigError('Invalid IP address in faucet_vips: %s' % err)
            for faucet_vip in self.faucet_vips:
                self.dyn_faucet_vips_by_ipv[faucet_vip.version].append(
                    faucet_vip)
            self.dyn_ipvs = list(self.dyn_faucet_vips_by_ipv.keys())
        if self.bgp_neighbor_addresses or self.bgp_neighbour_addresses:
            neigh_addresses = set(self.bgp_neighbor_addresses + self.bgp_neighbour_addresses)
            try:
                self.bgp_neighbor_addresses = [
                    ipaddress.ip_address(btos(ip)) for ip in neigh_addresses]
            except (ValueError, AttributeError, TypeError) as err:
                raise InvalidConfigError('Invalid IP address in bgp_neighbor_addresses: %s' % err)
            for bgp_neighbor_address in self.bgp_neighbor_addresses:
                self.dyn_bgp_neighbor_addresses_by_ipv[bgp_neighbor_address.version].append(
                    bgp_neighbor_address)
        if self.bgp_server_addresses:
            try:
                self.bgp_server_addresses = [
                    ipaddress.ip_address(btos(ip)) for ip in self.bgp_server_addresses]
            except (ValueError, AttributeError, TypeError) as err:
                raise InvalidConfigError('Invalid IP address in bgp_server_addresses: %s' % err)
            for bgp_server_address in self.bgp_server_addresses:
                self.dyn_bgp_server_addresses_by_ipv[bgp_server_address.version].append(
                    bgp_server_address)
            self.dyn_bgp_ipvs = list(self.dyn_bgp_server_addresses_by_ipv.keys())

        if self.bgp_as:
            test_config_condition(not isinstance(self.bgp_port, int), (
                'BGP port must be %s not %s' % (int, type(self.bgp_port))))
            test_config_condition(self.bgp_connect_mode not in ('active', 'passive', 'both'), (
                '%s must be active, passive or both' % self.bgp_connect_mode))
            test_config_condition(not ipaddress.IPv4Address(btos(self.bgp_routerid)), (
                '%s is not a valid IPv4 address' % (self.bgp_routerid)))
            test_config_condition(not self.bgp_neighbor_as, 'No BGP neighbor AS')
            test_config_condition(not self.bgp_neighbor_addresses, 'No BGP neighbor addresses')
            neighbor_ips = self.bgp_neighbor_addresses
            test_config_condition(len(neighbor_ips) != len(self.bgp_neighbor_addresses), (
                'Neighbor IPs is not the same length as BGP neighbor addresses'))
            peer_versions = [ip.version for ip in neighbor_ips]
            test_config_condition(len(peer_versions) != 1 and self.bgp_connect_mode != 'active', (
                'if using multiple address families bgp_connect_mode must be active'))

        if self.routes:
            try:
                self.routes = [route['route'] for route in self.routes]
                for route in self.routes:
                    try:
                        ip_gw = ipaddress.ip_address(btos(route['ip_gw']))
                        ip_dst = ipaddress.ip_network(btos(route['ip_dst']))
                    except (ValueError, AttributeError, TypeError) as err:
                        raise InvalidConfigError('Invalid IP address in route: %s' % err)
                    test_config_condition(ip_gw.version != ip_dst.version, 'ip_gw version does not match the ip_dst version')
                    self.add_route(ip_dst, ip_gw)
            except KeyError as err:
                raise InvalidConfigError('missing route config %s' % err)
            except TypeError:
                raise InvalidConfigError('%s is not a valid routes value' % self.routes)
        test_config_condition(self.acl_in and self.acls_in, 'found both acl_in and acls_in, use only acls_in')
        if self.acl_in and not isinstance(self.acl_in, list):
            self.acls_in = [self.acl_in,]
            self.acl_in = None
        if self.acls_in:
            for acl in self.acls_in:
                test_config_condition(not isinstance(acl, (int, str)), 'acl names must be int or str')

    @staticmethod
    def vid_valid(vid):
        """Return True if VID valid."""
        if isinstance(vid, int) and vid >= valve_of.MIN_VID and vid <= valve_of.MAX_VID:
            return True
        return False

    def reset_caches(self):
        """Reset dynamic caches."""
        self.dyn_host_cache = {}
        self.dyn_host_cache_by_port = {}
        self.dyn_neigh_cache_by_ipv = collections.defaultdict(dict)

    def reset_ports(self, ports):
        self.tagged = [port for port in ports if self in port.tagged_vlans]
        self.untagged = [port for port in ports if self == port.native_vlan]

    def add_cache_host(self, eth_src, port, cache_time):
        existing_entry = self.cached_host(eth_src)
        if existing_entry is not None:
            self.dyn_host_cache_by_port[existing_entry.port.number].remove(
                existing_entry)
        entry = HostCacheEntry(eth_src, port, cache_time)
        if port.number not in self.dyn_host_cache_by_port:
            self.dyn_host_cache_by_port[port.number] = set()
        self.dyn_host_cache_by_port[port.number].add(entry)
        self.dyn_host_cache[eth_src] = entry

    def expire_cache_host(self, eth_src):
        """Expire a host from caches."""
        entry = self.cached_host(eth_src)
        if entry is not None:
            self.dyn_host_cache_by_port[entry.port.number].remove(entry)
            del self.dyn_host_cache[eth_src]

    def cached_hosts_on_port(self, port):
        """Return all hosts learned on a port."""
        if port.number in self.dyn_host_cache_by_port:
            return list(self.dyn_host_cache_by_port[port.number])
        return []

    def cached_hosts_count_on_port(self, port):
        """Return count of all hosts learned on a port."""
        hosts_count = 0
        if port.number in self.dyn_host_cache_by_port:
            hosts_count = len(self.dyn_host_cache_by_port[port.number])
        return hosts_count

    def cached_host(self, eth_src):
        if eth_src in self.dyn_host_cache:
            return self.dyn_host_cache[eth_src]
        return None

    def cached_host_on_port(self, eth_src, port):
        """Return host cache entry if host in cache and on specified port."""
        entry = self.cached_host(eth_src)
        if entry and port == entry.port:
            return entry
        return None

    def clear_cache_hosts_on_port(self, port):
        """Clear all hosts learned on a port."""
        for entry in self.cached_hosts_on_port(port):
            self.expire_cache_host(entry.eth_src)

    def expire_cache_hosts(self, now, learn_timeout):
        """Expire stale host entries."""
        expired_hosts = []

        if self.dyn_oldest_host_time is None or now - self.dyn_oldest_host_time > learn_timeout:
            min_cache_time = now - learn_timeout
            self.dyn_oldest_host_time = now
            for entry in list(self.dyn_host_cache.values()):
                if (not entry.port.permanent_learn and entry.cache_time < min_cache_time):
                    expired_hosts.append(entry)
                else:
                    self.dyn_oldest_host_time = min(entry.cache_time, self.dyn_oldest_host_time)
            for entry in expired_hosts:
                self.expire_cache_host(entry.eth_src)
        return expired_hosts

    def ipvs(self):
        """Return list of IP versions configured on this VLAN."""
        return self.dyn_ipvs

    def bgp_ipvs(self):
        """Return list of IP versions for BGP configured on this VLAN."""
        return self.dyn_bgp_ipvs

    def faucet_vips_by_ipv(self, ipv):
        """Return list of VIPs with specified IP version on this VLAN."""
        return self.dyn_faucet_vips_by_ipv[ipv]

    def bgp_neighbor_addresses_by_ipv(self, ipv):
        """Return list of BGP neighbor addresses with specified IP version on this VLAN."""
        return self.dyn_bgp_neighbor_addresses_by_ipv[ipv]

    def bgp_server_addresses_by_ipv(self, ipv):
        """Return list of BGP server addresses with specified IP version on this VLAN."""
        return self.dyn_bgp_server_addresses_by_ipv[ipv]

    def routes_by_ipv(self, ipv):
        """Return route table for specified IP version on this VLAN."""
        return self.dyn_routes_by_ipv[ipv]

    def route_count_by_ipv(self, ipv):
        """Return route table count for specified IP version on this VLAN."""
        return len(self.dyn_routes_by_ipv[ipv])

    def add_route(self, ip_dst, ip_gw):
        """Add an IP route."""
        self.dyn_routes_by_ipv[ip_gw.version][ip_dst] = ip_gw
        if ip_gw not in self.dyn_gws_by_ipv[ip_gw.version]:
            self.dyn_gws_by_ipv[ip_gw.version][ip_gw] = set()
        self.dyn_gws_by_ipv[ip_gw.version][ip_gw].add(ip_dst)

    def del_route(self, ip_dst):
        """Delete an IP route."""
        ip_gw = self.dyn_routes_by_ipv[ip_dst.version][ip_dst]
        del self.dyn_routes_by_ipv[ip_dst.version][ip_dst]
        self.dyn_gws_by_ipv[ip_gw.version][ip_gw].remove(ip_dst)
        if not self.dyn_gws_by_ipv[ip_gw.version][ip_gw]:
            del self.dyn_gws_by_ipv[ip_gw.version][ip_gw]

    def ip_dsts_for_ip_gw(self, ip_gw):
        """Return list of IP destinations, for specified gateway."""
        if ip_gw in self.dyn_gws_by_ipv[ip_gw.version]:
            return list(self.dyn_gws_by_ipv[ip_gw.version][ip_gw])
        return []

    def all_ip_gws(self, ipv):
        """Return list of all IP gateways for specified IP version."""
        return list(self.dyn_gws_by_ipv[ipv].keys())

    def neigh_cache_by_ipv(self, ipv):
        """Return neighbor cache for specified IP version on this VLAN."""
        return self.dyn_neigh_cache_by_ipv[ipv]

    def neigh_cache_count_by_ipv(self, ipv):
        """Return number of hosts in neighbor cache for specified IP version on this VLAN."""
        return len(self.neigh_cache_by_ipv(ipv))

    def hosts_count(self):
        """Return number of hosts learned on this VLAN."""
        return len(self.dyn_host_cache)

    def __str__(self):
        port_list = [str(x) for x in self.get_ports()]
        ports = ','.join(port_list)
        return 'VLAN %s vid:%s ports:%s' % (self.name, self.vid, ports)

    def __repr__(self):
        return self.__str__()

    def get_ports(self):
        """Return list of all ports on this VLAN."""
        return list(self.tagged) + list(self.untagged)

    def hairpin_ports(self):
        """Return all ports with hairpin enabled."""
        return [port for port in self.get_ports() if port.hairpin]

    def mirrored_ports(self):
        """Return list of ports that are mirrored on this VLAN."""
        return [port for port in self.get_ports() if port.mirror]

    def lags(self):
        """Return dict of LAGs mapped to member ports."""
        lacp_ports = [port for port in self.get_ports() if port.lacp]
        lags = collections.defaultdict(list)
        for port in lacp_ports:
            lags[port.lacp].append(port)
        return lags

    def flood_ports(self, configured_ports, exclude_unicast):
        if exclude_unicast:
            return [port for port in configured_ports if port.unicast_flood]
        return configured_ports

    def tagged_flood_ports(self, exclude_unicast):
        return self.flood_ports(self.tagged, exclude_unicast)

    def untagged_flood_ports(self, exclude_unicast):
        return self.flood_ports(self.untagged, exclude_unicast)

    def output_port(self, port, hairpin=False):
        actions = port.mirror_actions()
        if self.port_is_untagged(port):
            actions.append(valve_of.pop_vlan())
        if hairpin:
            actions.append(valve_of.output_port(valve_of.OFP_IN_PORT))
        else:
            actions.append(valve_of.output_port(port.number))
        return actions

    def pkt_out_port(self, packet_builder, port, *args):
        vid = None
        if self.port_is_tagged(port):
            vid = self.vid
        pkt = packet_builder(vid, *args)
        return valve_of.packetout(port.number, pkt.data)

    def flood_pkt(self, packet_builder, random_order, *args):
        ofmsgs = []
        for vid, ports in (
                (self.vid, self.tagged_flood_ports(False)),
                (None, self.untagged_flood_ports(False))):
            if ports:
                pkt = packet_builder(vid, *args)
                flood_ofmsgs = [valve_of.packetout(port.number, pkt.data) for port in ports if port.running()]
                if random_order:
                    random.shuffle(flood_ofmsgs)
                ofmsgs.extend(flood_ofmsgs)
        return ofmsgs

    def port_is_tagged(self, port):
        """Return True if port number is an tagged port on this VLAN."""
        return port in self.tagged

    def port_is_untagged(self, port):
        """Return True if port number is an untagged port on this VLAN."""
        return port in self.untagged

    def is_faucet_vip(self, ipa):
        """Return True if IP is a VIP on this VLAN."""
        for faucet_vip in self.faucet_vips_by_ipv(ipa.version):
            if ipa == faucet_vip.ip:
                return True
        return False

    def ip_in_vip_subnet(self, ipa):
        """Return faucet_vip if IP in same IP network as a VIP on this VLAN."""
        for faucet_vip in self.faucet_vips_by_ipv(ipa.version):
            if ipa in faucet_vip.network:
                if ipa not in (
                        faucet_vip.network.network_address,
                        faucet_vip.network.broadcast_address):
                    return faucet_vip
        return None

    def from_connected_to_vip(self, src_ip, dst_ip):
        """Return True if src_ip in connected network and dst_ip is a VIP.

        Args:
            src_ip (ipaddress.ip_address): source IP.
            dst_ip (ipaddress.ip_address): destination IP
        Returns:
            True if local traffic for a VIP.
        """
        if self.is_faucet_vip(dst_ip) and self.ip_in_vip_subnet(src_ip):
            return True
        return False

    def to_conf(self):
        result = super(VLAN, self).to_conf()
        if result is not None:
            if self.routes:
                result['routes'] = [{'route': route} for route in self.routes]
            if self.faucet_vips:
                result['faucet_vips'] = [str(vip) for vip in self.faucet_vips]
            if self.bgp_neighbor_addresses:
                result['bgp_neighbor_addresses'] = [str(vip) for vip in self.bgp_neighbor_addresses]
            if self.bgp_server_addresses:
                result['bgp_server_addresses'] = [str(vip) for vip in self.bgp_server_addresses]
            if 'bgp_neighbor_as' in result:
                del result['bgp_neighbor_as']
            if 'bgp_neighbor_addresses' in result:
                del result['bgp_neighbor_addresses']
        return result



"""BGP implementation for FAUCET."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2018 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ipaddress

import eventlet
eventlet.monkey_patch()

from ryu.lib import hub # pylint: disable=wrong-import-position

from beka.beka import Beka # pylint: disable=wrong-import-position

class FaucetBgp(object):
    """Wrapper for Ryu BGP speaker."""

    def __init__(self, logger, metrics, send_flow_msgs):
        self.logger = logger
        self.metrics = metrics
        self._send_flow_msgs = send_flow_msgs
        self._dp_bgp_speakers = {}
        self._valves = None

    @staticmethod
    def _neighbor_states(bgp_speaker):
        """Return state of each neighbor for a BGP speaker as a list."""
        neighbor_states = []
        if bgp_speaker is not None:
            neighbor_states = bgp_speaker.neighbor_states()
        return neighbor_states

    def _bgp_up_handler(self, remote_ip, remote_as):
        self.logger.info('BGP peer router ID %s AS %s up' % (remote_ip, remote_as))

    def _bgp_down_handler(self, remote_ip, remote_as):
        self.logger.info('BGP peer router ID %s AS %s down' % (remote_ip, remote_as))

    def _bgp_route_handler(self, path_change, dp_id, vlan_vid):
        """Handle a BGP change event.

        Args:
            path_change (ryu.services.protocols.bgp.bgpspeaker.EventPrefix): path change
            dp_id (int): Datapath ID this path change was received for.
            vlan_vid (vlan_vid): VLAN VID this path change was received for.
        """
        if not self._valves or dp_id not in self._valves:
            return
        valve = self._valves[dp_id]
        if not vlan_vid in valve.dp.vlans:
            return

        vlan = valve.dp.vlans[vlan_vid]
        prefix = ipaddress.ip_network(str(path_change.prefix))

        if path_change.next_hop:
            nexthop = ipaddress.ip_address(str(path_change.next_hop))

            if vlan.is_faucet_vip(nexthop):
                self.logger.error(
                    'BGP nexthop %s for prefix %s cannot be us',
                    nexthop, prefix)
                return
            if vlan.ip_in_vip_subnet(nexthop) is None:
                self.logger.error(
                    'BGP nexthop %s for prefix %s is not a connected network',
                    nexthop, prefix)
                return

        flowmods = []
        if path_change.is_withdraw:
            self.logger.info(
                'BGP withdraw %s', prefix)
            flowmods = valve.del_route(vlan, prefix)
        else:
            self.logger.info(
                'BGP add %s nexthop %s', prefix, nexthop)
            flowmods = valve.add_route(vlan, nexthop, prefix)
        if flowmods:
            self._send_flow_msgs(valve, flowmods)

    @staticmethod
    def _bgp_vlans(valves):
        bgp_vlans = set()
        if valves:
            for valve in list(valves.values()):
                for vlan in valve.dp.bgp_vlans():
                    bgp_vlans.add(vlan)
        return bgp_vlans

    @staticmethod
    def _vlan_prefixes_by_ipv(vlan, ipv):
        vlan_prefixes = []
        for faucet_vip in vlan.faucet_vips_by_ipv(ipv):
            vlan_prefixes.append((str(faucet_vip), str(faucet_vip.ip)))
        routes = vlan.routes_by_ipv(ipv)
        for ip_dst, ip_gw in list(routes.items()):
            vlan_prefixes.append((str(ip_dst), str(ip_gw)))
        return vlan_prefixes

    def _create_bgp_speaker_for_vlan(self, vlan, dp_id, vlan_vid, ipv):
        """Set up BGP speaker for an individual VLAN if required.

        Args:
            vlan (valve VLAN): VLAN for BGP speaker.
            dp_id (int): Datapath ID for BGP speaker.
            vlan_vid (vlan_vid): VLAN VID for BGP speaker.
        Returns:
            ryu.services.protocols.bgp.bgpspeaker.BGPSpeaker: BGP speaker.
        """
        handler = lambda x: self._bgp_route_handler(x, dp_id, vlan_vid)

        if len(vlan.bgp_server_addresses_by_ipv(ipv)) > 1:
            self.logger.warning(
                'Only one server per IPv per vlan supported')
        server_address = sorted(vlan.bgp_server_addresses_by_ipv(ipv))[0]
        beka = Beka(
            local_address=str(server_address),
            bgp_port=vlan.bgp_port,
            local_as=vlan.bgp_as,
            router_id=vlan.bgp_routerid,
            peer_up_handler=self._bgp_up_handler,
            peer_down_handler=self._bgp_down_handler,
            route_handler=handler,
            error_handler=self.logger.warning
        )
        for ip_dst, ip_gw in self._vlan_prefixes_by_ipv(vlan, ipv):
            beka.add_route(prefix=str(ip_dst), next_hop=str(ip_gw))
        for bgp_neighbor_address in vlan.bgp_neighbor_addresses_by_ipv(ipv):
            beka.add_neighbor(
                connect_mode="passive",
                peer_ip=str(bgp_neighbor_address),
                peer_as=vlan.bgp_neighbor_as
                )

        hub.spawn(beka.run)
        return beka

    def shutdown_bgp_speakers(self):
        """Shutdown any active BGP speakers."""
        for vlan_bgp_speakers_by_ipv in list(self._dp_bgp_speakers.values()):
            for vlan_bgp_speakers in list(vlan_bgp_speakers_by_ipv.values()):
                for bgp_speaker in list(vlan_bgp_speakers.values()):
                    bgp_speaker.shutdown()
        self._dp_bgp_speakers = {}

    def reset(self, valves):
        """Set up a BGP speaker for every VLAN that requires it."""
        # TODO: port status changes should cause us to withdraw a route.
        # TODO: BGP speaker library does not cleanly handle del/add of same peer
        # TODO: BGP speaker can listen only on one address family at once
        # TODO: Ryu BGP supports only one speaker
        # (https://sourceforge.net/p/ryu/mailman/message/32699012/)
        # TODO: Ryu BGP cannot be restarted cleanly (so config cannot change at runtime)
        self._valves = valves
        if self._dp_bgp_speakers:
            self.logger.warning(
                'not updating existing BGP speaker, runtime BGP changes not supported')
            return
        bgp_vlans = self._bgp_vlans(self._valves)
        if not bgp_vlans:
            return
        if len(bgp_vlans) > 1:
            self.logger.warning(
                'only one BGP VLAN currently supported')
        bgp_vlan = sorted(bgp_vlans)[0]
        dp_id = bgp_vlan.dp_id
        vlan_vid = bgp_vlan.vid
        self._dp_bgp_speakers[dp_id] = {}
        for ipv in bgp_vlan.bgp_ipvs():
            self._dp_bgp_speakers[dp_id][ipv] = {
                vlan_vid: self._create_bgp_speaker_for_vlan(bgp_vlan, dp_id, vlan_vid, ipv)}

    def update_metrics(self, _now):
        """Update BGP metrics."""
        for dp_id, bgp_speakers_by_ipv in list(self._dp_bgp_speakers.items()):
            valve = self._valves[dp_id]
            for ipv, bgp_speakers in list(bgp_speakers_by_ipv.items()):
                for vlan_vid, bgp_speaker in list(bgp_speakers.items()):
                    # TODO: VLAN deconfigured, online reconfiguration while BGP active not supported.
                    if vlan_vid not in valve.dp.vlans:
                        continue
                    vlan = valve.dp.vlans[vlan_vid]
                    neighbor_states = self._neighbor_states(bgp_speaker)
                    for neighbor, neighbor_state in neighbor_states:
                        neighbor_labels = dict(
                            valve.base_prom_labels, vlan=vlan.vid, neighbor=neighbor)
                        self.metrics.bgp_neighbor_uptime_seconds.labels( # pylint: disable=no-member
                            **neighbor_labels).set(neighbor_state['info']['uptime'])
                        self.metrics.bgp_neighbor_routes.labels( # pylint: disable=no-member
                            **dict(neighbor_labels, ipv=ipv)).set(vlan.route_count_by_ipv(ipv))



"""Unit tests for DP"""

import unittest

from faucet.dp import DP

class FaucetDPConfigTest(unittest.TestCase): # pytype: disable=module-attr
    """Test that DP serialises config as it receives it"""

    def setUp(self):
        """Defines the default config - this should match the documentation"""

        self.default_config = {
            'advertise_interval': 30,
            'arp_neighbor_timeout': 250,
            'combinatorial_port_flood': False,
            'cookie': 1524372928,
            'drop_broadcast_source_address': True,
            'drop_lldp': True,
            'drop_spoofed_faucet_mac': True,
            'faucet_dp_mac': '0e:00:00:00:00:01',
            'group_table': False,
            'group_table_routing': False,
            'hardware': 'Open vSwitch',
            'high_priority': 9001,
            'highest_priority': 9099,
            'ignore_learn_ins': 10,
            'interface_ranges': {},
            'interfaces': {},
            'learn_ban_timeout': 10,
            'learn_jitter': 10,
            'lldp_beacon': {},
            'low_priority': 9000,
            'lowest_priority': 0,
            'max_host_fib_retry_count': 10,
            'max_hosts_per_resolve_cycle': 5,
            'max_resolve_backoff_time': 32,
            'metrics_rate_limit_sec': 0,
            'ofchannel_log': None,
            'packetin_pps': None,
            'pipeline_config_dir': '/etc/faucet',
            'priority_offset': 0,
            'proactive_learn': True,
            'stack': None,
            'timeout': 300,
            'use_idle_timeout': False
        }

    def test_basic_config(self):
        """Tests the minimal config"""
        dp_id = 12345

        input_config = {
            'interfaces': {
                1: {}
            }
        }
        output_config = {
            'description': str(dp_id),
            'dp_id': dp_id,
            'interfaces': {}
        }

        expected_config = self.default_config
        expected_config.update(input_config)
        expected_config.update(output_config)

        dp = DP(dp_id, None, input_config) # pylint: disable=invalid-name
        output_config = dp.to_conf()

        self.assertEqual(output_config, expected_config)

        key_exceptions = [
            'lldp_beacon_ports',
            'output_only_ports',
            'vlans',
            'routers',
            'acls',
            'ports',
            'stack_ports',
            '_id',
            'groups',
            'name'
        ]
        dict_keys = set(dp.__dict__.keys())
        conf_keys = set(dp.to_conf().keys())

        for exception in key_exceptions:
            dict_keys.remove(exception)

        self.assertEqual(dict_keys, conf_keys)



"""Standard FAUCET pipeline."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2018 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pipeline definition is a list of tuples, one tuple per table,
# starting at OpenFlow table 0.
# The first item in tuple is the table name.
# The second item in the tuple is a tuple of OpenFlow matches, that table ues.
FAUCET_PIPELINE = (
    ('port_acl', None),
    ('vlan', ('eth_dst', 'eth_type', 'in_port', 'vlan_vid')),
    ('vlan_acl', None),
    ('eth_src', ('eth_dst', 'eth_src', 'eth_type', 'in_port', 'vlan_vid')),
    ('ipv4_fib', ('eth_type', 'ipv4_dst', 'vlan_vid')),
    ('ipv6_fib', ('eth_type', 'ipv6_dst', 'vlan_vid')),
    ('vip', ('arp_tpa', 'eth_dst', 'eth_type', 'icmpv6_type', 'ip_proto')),
    ('eth_dst', ('eth_dst', 'in_port', 'vlan_vid')),
    ('flood', ('eth_dst', 'in_port', 'vlan_vid')))



#!/usr/bin/env python3

"""Shows the crash in the faucet log produced by given input"""

import logging
import os
import sys
from faucet import faucet
from ryu.controller import dpset
from faucet import faucet_experimental_api
import Fake

def main():
    # go through all files in directory
    # read file and store in string
    with open(sys.argv[1]) as pkt:
        packet_data = pkt.read()

    # start faucet
    application = faucet.Faucet(dpset=dpset.DPSet(), faucet_experimental_api=faucet_experimental_api.FaucetExperimentalAPI())
    application.start()

    # make sure dps are running
    for dp_id, valve in list(application.valves.items()):
        valve.dp.running = True

    # create data from read file
    byte_data = None
    try:
        byte_data = bytearray.fromhex(packet_data)
    except (ValueError, TypeError):
        pass

    if byte_data is not None:
        # create fake packet
        dp = Fake.Datapath(1)
        msg = Fake.Message(datapath=dp, cookie=1524372928, port=1, data=byte_data, in_port=1)
        pkt = Fake.RyuEvent(msg)

        # send packet to faucet and display error produced
        application.packet_in_handler(pkt)

if __name__ == "__main__":
    # make sure user specifies the afl crash folder
    if len(sys.argv) == 2:
        main()
    else:
        print('USAGE: python3 display_packet_crash.py <AFL_CRASH_FILE>')

os._exit(0)



#!/usr/bin/env python
#
# (c) Copyright 2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Convert output of ovs-ofctl/dpctl dump-flows into pipeline JSON which
    would support those flows on an HPE Aruba switch. If the set of flows
    cannot fit into a possible pipeline within the HPE Aruba switch, one
    or more errors will be generated to indicate the reason. """

import sys
import re
import os
import json

# =========================== Match Field Data ================================

# This tool is coded according to the "Flow Syntax" section of the following
# OVS documentation:
#   http://openvswitch.org/support/dist-docs/ovs-ofctl.8.txt

# Match fields are keyed to the group they correspond with.
# Those fields which are their own group are given high numbers.
OVS_MATCH_FIELDS = {
    # Basic OF match fields
    'in_port':4, 'dl_vlan':4, 'dl_vlan_pcp':4,
    'dl_src':1, 'dl_dst':1, 'dl_type':1,
    'nw_src':2, 'nw_dst':3,
    'nw_proto':4, 'ip_proto':4, 'ip_dscp':4,
    'tcp_src':4, 'tcp_dst':4, 'udp_src':4, 'udp_dst':4,
    'tcp_flags':899, 'icmp_type':900, 'icmp_code':901,
    # Nicira extensions
    'arp_op':902, 'arp_spa':903, 'arp_tpa':904, 'arp_sha':905, 'arp_tha':906,
    'ipv6_src':2, 'ipv6_dst':3, 'ipv6_label':907, 'nd_target':908}

# Match fields which are not supported by HPE Aruba switches,
# regardless of the pipeline configuration.
OVS_UNSUPPORTED_MATCH = [
    # Basic OF match fields
    'nw_tos', 'nw_ttl', 'nw_ecn', 'ip_ecn',
    'sctp_src', 'sctp_dst', 'metadata',
    # Nicira extensions
    'vlan_tci', 'ip_frag', 'nd_sll', 'nd_tll',
    'mpls_bos', 'mpls_label', 'mpls_tc',
    'tun_id', 'tunnel_id', 'tun_flags', 'tun_src', 'tun_dst',
    'tun_ipv6_src', 'tun_ipv6_dst', 'tun_gbp_id', 'tun_gbp_flags',
    'tun_metadataidx', 'regidx', 'xregidx', 'pkt_mark', 'actset_output',
    'conj_id', 'ct_state', 'ct_zone', 'ct_mark', 'ct_label']

OVS_KNOWN_MATCH = set(OVS_MATCH_FIELDS.keys()).union(set(OVS_UNSUPPORTED_MATCH))

# OVS match abbreviations, keyed to what they abbreviate
OVS_MATCH_ABBREV = {
    'ip': 'dl_type=0x0800',
    'ipv6': 'dl_type=0x86dd',
    'icmp': 'dl_type=0x0800,nw_proto=1',
    'icmp6': 'dl_type=0x86dd,nw_proto=58',
    'tcp': 'dl_type=0x0800,nw_proto=6',
    'tcp6': 'dl_type=0x86dd,nw_proto=6',
    'udp': 'dl_type=0x0800,nw_proto=17',
    'udp6': 'dl_type=0x86dd,nw_proto=17',
    'sctp': 'dl_type=0x0800,nw_proto=132',
    'sctp6': 'dl_type=0x86dd,nw_proto=132',
    'arp': 'dl_type=0x0806',
    'rarp': 'dl_type=0x8035',
    'mpls': 'dl_type=0x8847',
    'mplsm': 'dl_type=0x8848',
    # Assume deprecated tp_src/dst are UDP (to avoid complicated parsing)
    'tp_src': 'udp_src',
    'tp_dst': 'udp_dst',
    # Special case use of vlan_tci to match packets without a VLAN tag,
    # to dl_vlan=0. Per OF1.3.3 spec p.117:
    #   - Testing for an exact match with 0x0 matches only packets without
    #   * an 802.1Q header.
    'vlan_tci=0x0000/0x1fff': 'dl_vlan=0'}

# Conversion from OVS field name to RYU field name. Keys should be identical
# to the keys in OVS_MATCH_FIELDS.
JSON_FIELDS = {
    # Basic OF match fields
    'in_port':'in_port', 'dl_vlan':'vlan_vid', 'dl_vlan_pcp':'vlan_pcp',
    'dl_src':'eth_src', 'dl_dst':'eth_dst', 'dl_type':'eth_type',
    'nw_src':'ipv4_src', 'nw_dst':'ipv4_dst',
    'nw_proto':'ip_proto', 'ip_proto':'ip_proto', 'ip_dscp':'ip_dscp',
    'tcp_src':'tcp_src', 'tcp_dst':'tcp_dst', 'udp_src':'udp_src', 'udp_dst':'udp_dst',
    'tcp_flags':'tcp_flags', 'icmp_type':'icmpv4_type', 'icmp_code':'icmpv4_code',
    # Nicira extensions
    'arp_op':'arp_op', 'arp_spa':'arp_spa', 'arp_tpa':'arp_tpa', 'arp_sha':'arp_sha', 'arp_tha':'arp_tha',
    'ipv6_src':'ipv6_src', 'ipv6_dst':'ipv6_dst', 'ipv6_label':'ipv6_flabel', 'nd_target':'ipv6_nd_target'}
if sorted(JSON_FIELDS.keys()) != sorted(OVS_MATCH_FIELDS.keys()):
    print('ERROR: Key mismatch between JSON_FIELDS and OVS_MATCH_FIELDS:\n')
    print(set(JSON_FIELDS.keys()).symmetric_difference(set(OVS_MATCH_FIELDS.keys())))
    exit(2)

# Fields which HPE Aruba supports as setfield in any pipeline (keyed with RYU field names, not OVS)
ARUBA_SETFIELDS = [
    'eth_dst', 'eth_src', 'vlan_vid', 'vlan_pcp', 'ip_dscp',
    'ipv4_src', 'ipv4_dst', 'tcp_src', 'tcp_dst', 'udp_src', 'udp_dst']

DEBUG = ('DEBUG' in os.environ)
GENERATE_JSON = True

# =========================== Utility Functions ================================
def debug(arg):
    if DEBUG:
        print(arg)

def error(arg):
    print(arg)
    global GENERATE_JSON
    GENERATE_JSON = False

# =========================== Input Processing ================================

# Check for the input data file
if len(sys.argv) < 2:
    print("Please specify a filename which contains the output of 'ovs-ofctl dump-flows'")
    exit(1)

# Allocate variables which will hold data extracted from OVS output
TABLE_MATCH = {} # Table ID key, value is a list of exact match keys
TABLE_WILDCARDS = {} # Table ID key, value is a list of wildcardable match keys
TABLE_MASKS = {} # Table ID key, value is a list of maskable match keys
TABLE_SIZE = {} # Table ID key, value is number of flows in table

# Iterate over all lines of the file, gathering data
debug("=== Per-flow pipeline analysis ===")
for line in open(sys.argv[1]):
    # Skip empty lines
    line = line.rstrip().lstrip()
    if not line:
        continue

    # debug("FLOW: "+line)
    line_data = re.split(r'\s+', line)
    errors = 0
    table = None
    match = None

    # Identify the table
    for data in line_data:
        if "table=" not in data:
            continue

        # Store the numeric table ID
        table = data.split('=')[1]
        table = re.sub(r'\D', '', table)

    # Identify match criteria
    for data in line_data:
        if "priority=" not in data:
            continue
        match = data

    # Verify that we found both table ID and match data
    if table is None:
        debug("Failed to identify table ID in line:\n "+line)
        continue
    if match is None:
        debug("Failed to identify match data in line:\n "+line)
        continue

    # Handle deprecated tp_src/dst interpretation, which is context-dependent
    if re.match('.*[^c]tp_(src|dst).*', match) != None:
        if "udp" in match or "proto=17" in match:
            match = re.sub(r'([^c]*)tp_(dst|src)', r'\1udp_\2', match)
        else:
            match = re.sub(r'([^c]*)tp_(dst|src)', r'\1tcp_\2', match)

    # Replace abbreviations
    for abbrev in OVS_MATCH_ABBREV:
        fields = match.split(',')
        fields = [re.sub("^"+abbrev+"$", OVS_MATCH_ABBREV[abbrev], f) for f in fields]
        match = ','.join(fields)

    # Get the list of matched and masked fields
    match_keys = set([])
    masks = set([])
    for m in match.split(','):
        # Skip empty match field
        if m == "":
            continue

        # Parse the match key
        mp = m.split('=')
        key = mp[0]
        match_keys.add(key)

        # Parse the match value
        if len(mp) >= 2:
            value = m.split('=')[1]
            if "/" in value:
                masks.add(key)
        else:
            error("Failed to parse special key-value abbreviation: "+str(m))

    match_keys.remove('priority') # Ignore priority, not a match field

    # Verify our hard-coded OVS match fields are complete
    if not OVS_KNOWN_MATCH.issuperset(match_keys):
        error("Attempted to match unknown field(s) "+str(list(match_keys.difference(OVS_KNOWN_MATCH)))+" in flow:\n "+line)
        errors += 1

    # Check unsupported matches
    for unsupp in OVS_UNSUPPORTED_MATCH:
        if unsupp in match_keys:
            error("Match field '"+unsupp+"' is not supported, but was used in:\n "+line)
            errors += 1

    # If we've hit errors, skip this flow because it will complicate the
    # global validation done later.
    if errors > 0:
        continue

    # Increment table size
    if table in TABLE_SIZE:
        TABLE_SIZE[table] += 1
    else:
        TABLE_SIZE[table] = 1
        TABLE_MATCH[table] = set(match_keys)
        TABLE_WILDCARDS[table] = set([])
        TABLE_MASKS[table] = set([])

    # Record match fields
    tm = TABLE_MATCH[table]
    exact = tm.intersection(match_keys)
    wildcard = tm.symmetric_difference(match_keys)
    TABLE_MATCH[table] = set(exact)
    TABLE_WILDCARDS[table] = set(wildcard).union(TABLE_WILDCARDS[table])
    TABLE_MASKS[table] = TABLE_MASKS[table].union(masks)

# Globals used in validation
debug("\n=== Global and per-table pipeline analysis ===")
MAX_SUPPORTED_TABLES = 12
MIN_TCAM_SIZE = 2
MIN_HASH_SIZE = 16
MAX_TCAM_TILES = 8 * 1024
MAX_HASH_TILES = 64 * 1024

# Check if too many tables were used
if len(TABLE_SIZE.keys()) > MAX_SUPPORTED_TABLES:
    error("HPE Aruba switches support a maximum of "+str(MAX_SUPPORTED_TABLES)+" tables, but "+str(len(TABLE_SIZE.keys()))+" were used:\n "+str(TABLE_SIZE.keys()))

# Check if table 0 (required) was used
if '0' not in TABLE_SIZE:
    error("Table 0 was not used, but is required by the OpenFlow specification")

# Get a numerically-sorted list of table IDs
tables = TABLE_SIZE.keys()
tables.sort(key=int)

# Display and analyze gathered data to check for unsupported conditions
tcam_tiles = 0
hash_tiles = 0
for table in tables:
    # Get all fields being matched
    size = TABLE_SIZE[table]
    exact = TABLE_MATCH[table]
    wildcard = TABLE_WILDCARDS[table]
    mask = TABLE_MASKS[table]
    all_matches = exact.union(wildcard).union(mask)

    # Special-case: If a table had flows but none of the flows specified
    # match criteria, we'll specify at least one wildcard (ETH_TYPE) so that
    # the table is considered a TCAM. A hash must have all match fields specified,
    # but since no flows specified match criteria we know a TCAM is expected.
    # Wildcard ETH_TYPE since it is a dependency of many other fields.
    if len(all_matches) == 0:
        TABLE_WILDCARDS[table].add('dl_type')
        wildcard = TABLE_WILDCARDS[table]
        all_matches = exact.union(wildcard).union(mask)

    # Display table data
    debug("TABLE #"+table+" has "+str(size)+" entries")
    if len(exact) > 0:
        debug("  exact-match: "+str(sorted(list(exact))))
    if len(wildcard) > 0:
        debug("  wildcards: "+str(sorted(list(wildcard))))
    if len(mask) > 0:
        debug("  maskable: "+str(sorted(list(mask))))

    # Determine number of match groups
    groups = [OVS_MATCH_FIELDS[m] for m in all_matches]
    groups = set(groups)
    gc = len(groups)

    # Automatically upconvert Hash->TCAM if attempting to match 4 groups in hash
    if len(wildcard) == 0 and len(mask) == 0 and gc == 4:
        wildcard = exact.copy()
        exact.clear()
        debug("  ** Table #"+table+" has been converted from Hash to TCAM, due to matching 4 groups")

    # Calculate resource usage, based on table type
    if len(wildcard) > 0 or len(mask) > 0:
        table_type = "TCAM"
        if gc > 4:
            error("Table #"+table+" attempts to match fields from "+str(gc)+" groups. Maximum of 4 match groups supported in "+table_type)

        if size < MIN_TCAM_SIZE:
            debug("  ** Table #"+table+" has been auto-resized to minimum size of "+str(MIN_TCAM_SIZE))
            size = MIN_TCAM_SIZE

        mult = gc if gc != 3 else 4  # TCAM: 3 groups use same as 4 groups
        tiles = size * mult
        tcam_tiles += tiles
    else:
        table_type = "Hash"
        if gc > 3:
            error("Table #"+table+" attempts to match fields from "+str(gc)+" groups. Maximum of 3 match groups supported in "+table_type)

        if size < MIN_HASH_SIZE:
            debug("  ** Table #"+table+" has been auto-resized to minimum size of "+str(MIN_HASH_SIZE))
            size = MIN_HASH_SIZE

        mult = gc if gc != 3 else 2  # Hash: 3 groups use same as 2 groups
        tiles = size * mult
        hash_tiles += tiles

    # Store any adjusted values
    TABLE_SIZE[table] = size
    TABLE_MATCH[table] = exact
    TABLE_WILDCARDS[table] = wildcard
    TABLE_MASKS[table] = mask

    # Calculate resource allocation for this table
    debug("  allocation: "+str(len(groups))+" groups "+str(list(groups))+" using "+str(tiles)+" "+table_type+" resources")

# Verify that tables will fit into available hardware resources
debug("Total resources:  TCAM={} ({:.2f}%)  Hash={} ({:.2f}%)".format(tcam_tiles, 100*float(tcam_tiles)/MAX_TCAM_TILES,
                                                                      hash_tiles, 100*float(hash_tiles)/MAX_HASH_TILES))
if tcam_tiles > MAX_TCAM_TILES:
    error("Pipeline uses "+str(tcam_tiles)+" TCAM resources. Maximum of "+str(MAX_TCAM_TILES)+" available.")
if hash_tiles > MAX_HASH_TILES:
    error("Pipeline uses "+str(hash_tiles)+" Hash resources. Maximum of "+str(MAX_HASH_TILES)+" available.")

# Exit now if not generating JSON
if not GENERATE_JSON:
    exit(1)

# Generate JSON for RYU pipeline format
debug("\n=== Auto-generated pipeline JSON ===")
prev_tables = set([])
JSON = '['
for table in tables:
    # Get all fields being matched
    size = TABLE_SIZE[table]
    exact = TABLE_MATCH[table]
    wildcard = TABLE_WILDCARDS[table]
    mask = TABLE_MASKS[table]
    all_matches = exact.union(wildcard).union(mask)

    # Table header information
    JSON += '{"max_entries": '+str(size)+','
    JSON += '"name": "Table '+table+'",'
    JSON += '"table_id": '+table+','
    JSON += '"metadata_match": 0,'
    JSON += '"metadata_write": 0,'
    JSON += '"config": 3,'
    JSON += '"properties": ['

    # Matches
    JSON += '{"type":8, "name":"OFPTFPT_MATCH", "oxm_ids": ['
    for m in all_matches:
        hasmask = ""
        if m in mask:
            hasmask = ' , "hasmask": true'
        JSON += '{ "type": "'+JSON_FIELDS[m]+'", "name": "'+JSON_FIELDS[m]+'"'+hasmask+' },'

    # Trim trailing common from last match
    if len(all_matches) > 0:
        JSON = JSON.rstrip(',')
    JSON += ']},'

    # Wildcards
    JSON += '{"type":10, "name": "OFPTFPT_WILDCARDS", "oxm_ids": ['
    for w in wildcard:
        JSON += '{ "type": "'+JSON_FIELDS[w]+'", "name":"'+JSON_FIELDS[w]+'" },'

    # Trim trailing common from last wildcard
    if len(wildcard) > 0:
        JSON = JSON.rstrip(',')
    JSON += ']},'

    # Now that we've generated the match+wildcard criteria, we can assume that
    # all other tables will support all actions, so generate the same action
    # criteria, regardless of what the flows actually tried to use.
    genericSetfields = ['{"type":"'+f+'","name":"'+f+'"}' for f in ARUBA_SETFIELDS]
    genericActions = [
        '{"type":0,"name":"OFPAT_OUTPUT"}',
        '{"type":17,"name":"OFPAT_PUSH_VLAN"}',
        '{"type":18,"name":"OFPAT_POP_VLAN"}',
        '{"type":22,"name":"OFPAT_GROUP"}',
        '{"type":23,"name":"OFPAT_SET_NW_TTL"}',
        '{"type":25,"name":"OFPAT_SET_FIELD"}']
    genericInstructions = [
        '{"type":1,"name":"OFPIT_GOTO_TABLE"}',
        '{"type":3,"name":"OFPIT_WRITE_ACTIONS"}',
        '{"type":4,"name":"OFPIT_APPLY_ACTIONS"}',
        '{"type":5,"name":"OFPIT_CLEAR_ACTIONS"}',
        '{"type":6,"name":"OFPIT_METER"}']
    genericProps = ','.join(
        ['{ "type":0, "name":"OFPTFPT_INSTRUCTIONS", "instruction_ids": [ '+','.join(genericInstructions)+' ] }',
         '{ "type":1, "name":"OFPTFPT_INSTRUCTIONS_MISS", "instruction_ids": [ '+','.join(genericInstructions)+' ] }',
         '{ "type":4, "name":"OFPTFPT_WRITE_ACTIONS", "action_ids": [ '+','.join(genericActions)+' ] }',
         '{ "type":5, "name":"OFPTFPT_WRITE_ACTIONS_MISS", "action_ids": [ '+','.join(genericActions)+' ] }',
         '{ "type":6, "name":"OFPTFPT_APPLY_ACTIONS", "action_ids": [ '+','.join(genericActions)+' ] }',
         '{ "type":7, "name":"OFPTFPT_APPLY_ACTIONS_MISS", "action_ids": [ '+','.join(genericActions)+' ] }',
         '{ "type":12, "name":"OFPTFPT_WRITE_SETFIELD", "oxm_ids": [ '+','.join(genericSetfields)+' ] }',
         '{ "type":13, "name":"OFPTFPT_WRITE_SETFIELD_MISS", "oxm_ids": [ '+','.join(genericSetfields)+' ] }',
         '{ "type":14, "name":"OFPTFPT_APPLY_SETFIELD", "oxm_ids": [ '+','.join(genericSetfields)+' ] }',
         '{ "type":15, "name":"OFPTFPT_APPLY_SETFIELD_MISS", "oxm_ids": [ '+','.join(genericSetfields)+' ] }'])

    # Remove GOTO from last table
    prev_tables.add(table)
    remaining_tables = set(TABLE_SIZE.keys()).difference(prev_tables)
    lastTable = (len(remaining_tables) == 0)
    if lastTable:
        genericProps = ''.join(genericProps.rsplit('{"type":1,"name":"OFPIT_GOTO_TABLE"}, ', 1))
    JSON += genericProps+','

    JSON += '{ "type":2, "name": "OFPTFPT_NEXT_TABLES", "table_ids": [ '+','.join(remaining_tables)+' ] },'
    JSON += '{ "type":3, "name": "OFPTFPT_NEXT_TABLES_MISS", "table_ids": [ '+','.join(remaining_tables)+' ] }'
    JSON += ']}'
    if not lastTable:
        JSON += ','

# Wrap things up and print ...
JSON += ']'

# Pretty-print the condensed JSON string, for easier diffs
jsonobj = json.loads(JSON)
json.dump(jsonobj, sys.stdout, sort_keys=True, indent=4, separators=(',', ': '))



"""Unit tests for VLAN"""

import unittest
from ipaddress import ip_address, ip_network, ip_interface

from faucet.vlan import VLAN


class FaucetVLANBaseTest(unittest.TestCase): # pytype: disable=module-attr
    """Set up defaults for VLAN tests"""

    def setUp(self):
        """Defines the default config - this should match the documentation"""

        self.default_config = {
            'acl_in': None,
            'acls_in': None,
            'bgp_as': None,
            'bgp_connect_mode': 'both',
            'bgp_local_address': None,
            'bgp_neighbour_addresses': [],
            'bgp_neighbour_as': None,
            'bgp_port': 9179,
            'bgp_routerid': None,
            'bgp_server_addresses': ['0.0.0.0', '::'],
            'description': None,
            'faucet_mac': '0e:00:00:00:00:01',
            'faucet_vips': [],
            'max_hosts': 255,
            'minimum_ip_size_check': True,
            'proactive_arp_limit': 2052,
            'proactive_nd_limit': 2052,
            'routes': None,
            'targeted_gw_resolution': False,
            'unicast_flood': True,
        }

class FaucetVLANConfigTest(FaucetVLANBaseTest):
    """Test that VLAN serialises config as it receives it"""

    default_config = None

    def test_basic_config(self):
        """Tests the minimal config"""

        input_config = {
            'vid': 100
        }

        expected_config = self.default_config
        expected_config.update(input_config)

        vlan = VLAN(1, 1, input_config)
        output_config = vlan.to_conf()

        self.assertEqual(output_config, expected_config)

        key_exceptions = [
            'name',
            'tagged',
            'dyn_gws_by_ipv',
            'dyn_host_cache_by_port',
            'dp_id',
            'bgp_neighbor_addresses',
            'bgp_neighbor_as',
            'dyn_routes_by_ipv',
            '_id',
            'dyn_neigh_cache_by_ipv',
            'dyn_ipvs',
            'dyn_bgp_ipvs',
            'dyn_host_cache',
            'dyn_faucet_vips_by_ipv',
            'dyn_bgp_neighbor_addresses_by_ipv',
            'dyn_bgp_server_addresses_by_ipv',
            'untagged'
        ]
        dict_keys = set(vlan.__dict__.keys())
        conf_keys = set(vlan.to_conf().keys())

        for exception in key_exceptions:
            dict_keys.remove(exception)

        self.assertEqual(dict_keys, conf_keys)

    def test_with_routes(self):
        """Tests a config with routes"""

        input_config = {
            'routes': [
                {'route' : {'ip_dst': '10.99.99.0/24', 'ip_gw': '10.0.0.1'}},
                {'route' : {'ip_dst': '10.99.98.0/24', 'ip_gw': '10.0.0.99'}}
            ],
            'vid': 100
        }

        expected_config = self.default_config
        expected_config.update(input_config)

        vlan = VLAN(1, 1, input_config)
        output_config = vlan.to_conf()

        self.assertEqual(output_config, expected_config)

    def test_with_vips(self):
        """Tests a config with virtual IPs"""

        input_config = {
            'faucet_vips': ['10.0.0.254/24'],
            'vid': 100
        }

        expected_config = self.default_config
        expected_config.update(input_config)

        vlan = VLAN(1, 1, input_config)
        output_config = vlan.to_conf()

        self.assertEqual(output_config, expected_config)

class FaucetVLANMethodTest(FaucetVLANBaseTest):
    """Initialises VLANs with different configs and sanity checks the associated methods"""

    def setUp(self):
        """Use the default config as a base"""

        super(FaucetVLANMethodTest, self).setUp()

        self.input_config = self.default_config

    def test_ipvs_no_ips(self):
        """Tests the ipvs() method with no vips"""

        vlan = VLAN(1, 1, self.input_config)
        self.assertEqual(vlan.ipvs(), [])

    def test_ipvs_ipv4(self):
        """Tests the ipvs() method with an IPv4 vip"""

        self.input_config.update({
            'faucet_vips': ['10.0.0.254/24']
        })

        vlan = VLAN(1, 1, self.input_config)
        self.assertEqual(vlan.ipvs(), [4])

    def test_ipvs_ipv6(self):
        """Tests the ipvs() method with an IPv6 vip"""

        self.input_config.update({
            'faucet_vips': ['2001::1/16']
        })

        vlan = VLAN(1, 1, self.input_config)
        self.assertEqual(vlan.ipvs(), [6])

    def test_ipvs_ipv4_ipv6(self):
        """Tests the ipvs() method with both IPv4 and IPv6 vips"""

        self.input_config.update({
            'faucet_vips': [
                '2001::1/16',
                'fe80::1/64',
                '10.0.0.254/24'
            ]
        })

        vlan = VLAN(1, 1, self.input_config)
        self.assertEqual(set(vlan.ipvs()), set([4, 6]))

    def test_bgp_servers_change_bgp_ipvs_ipv4(self):
        """Tests the ipvs() method with an IPv4 BGP server"""

        self.input_config.update({
            'bgp_server_addresses': ['127.0.0.1']
        })

        vlan = VLAN(1, 1, self.input_config)
        self.assertEqual(vlan.bgp_ipvs(), [4])

    def test_bgp_servers_change_bgp_ipvs_ipv6(self):
        """Tests the ipvs() method with an IPv4 BGP server"""

        self.input_config.update({
            'bgp_server_addresses': ['::1']
        })

        vlan = VLAN(1, 1, self.input_config)
        self.assertEqual(vlan.bgp_ipvs(), [6])

    def test_bgp_servers_change_bgp_ipvs_both(self):
        """Tests the ipvs() method with an IPv4 BGP server"""

        self.input_config.update({
            'bgp_server_addresses': ['127.0.0.1', '::1']
        })

        vlan = VLAN(1, 1, self.input_config)
        self.assertEqual(vlan.bgp_ipvs(), [4, 6])
        self.assertEqual(vlan.bgp_server_addresses_by_ipv(4), [ip_address('127.0.0.1')])
        self.assertEqual(vlan.bgp_server_addresses_by_ipv(6), [ip_address('::1')])

    def test_faucet_vips_by_ipv_none(self):
        """Tests the faucet_vips_by_ipv() method when there are no vips"""

        vlan = VLAN(1, 1, self.input_config)
        self.assertEqual(vlan.faucet_vips_by_ipv(4), [])
        self.assertEqual(vlan.faucet_vips_by_ipv(6), [])

    def test_faucet_vips_by_ipv_both(self):
        """Tests the faucet_vips_by_ipv() method when there are both IPv4 and IPv6 vips"""

        self.input_config.update({
            'faucet_vips': [
                '2001::1/16',
                'fe80::1/64',
                '10.0.0.254/24'
            ]
        })

        vlan = VLAN(1, 1, self.input_config)
        self.assertEqual(set(vlan.faucet_vips_by_ipv(4)), set([
            ip_interface('10.0.0.254/24')
        ]))
        self.assertEqual(set(vlan.faucet_vips_by_ipv(6)), set([
            ip_interface('2001::1/16'),
            ip_interface('fe80::1/64')
        ]))

    def test_routes_by_ipv_none(self):
        """Tests the routes_by_ipv() and route_count_by_ipv() methods with no routes"""

        vlan = VLAN(1, 1, self.input_config)
        self.assertEqual(vlan.routes_by_ipv(4), {})
        self.assertEqual(vlan.routes_by_ipv(6), {})
        self.assertEqual(vlan.route_count_by_ipv(4), 0)
        self.assertEqual(vlan.route_count_by_ipv(6), 0)

    def test_routes_by_ipv_both(self):
        """Tests the routes_by_ipv() and route_count_by_ipv() methods with both
        IPv4 and IPv6 routes"""

        self.input_config.update({
            'routes': [
                {'route': {'ip_dst': '10.99.99.0/24', 'ip_gw': '10.0.0.1'}},
                {'route': {'ip_dst': '10.99.98.0/24', 'ip_gw': '10.0.0.99'}},
                {'route': {'ip_dst': '10.99.97.0/24', 'ip_gw': '10.0.0.99'}},
                {'route': {'ip_dst': 'fc00::10:0/112', 'ip_gw': 'fc00::1:1'}},
                {'route': {'ip_dst': 'fc00::20:0/112', 'ip_gw': 'fc00::1:99'}}
            ],
        })

        vlan = VLAN(1, 1, self.input_config)

        self.assertEqual(vlan.routes_by_ipv(4), {
            ip_network('10.99.99.0/24'): ip_address('10.0.0.1'),
            ip_network('10.99.98.0/24'): ip_address('10.0.0.99'),
            ip_network('10.99.97.0/24'): ip_address('10.0.0.99'),
        })
        self.assertEqual(vlan.routes_by_ipv(6), {
            ip_network('fc00::10:0/112'): ip_address('fc00::1:1'),
            ip_network('fc00::20:0/112'): ip_address('fc00::1:99'),
        })
        self.assertEqual(vlan.route_count_by_ipv(4), 3)
        self.assertEqual(vlan.route_count_by_ipv(6), 2)

    def test_modify_routes_v4(self):
        """Tests the add_route() and remove_route() methods with IPv4 routes"""

        vlan = VLAN(1, 1, self.input_config)

        self.assertEqual(vlan.routes_by_ipv(4), {})
        vlan.add_route(ip_network('10.99.99.0/24'), ip_address('10.0.0.1'))
        vlan.add_route(ip_network('10.99.98.0/24'), ip_address('10.0.0.99'))
        self.assertEqual(vlan.routes_by_ipv(4), {
            ip_network('10.99.99.0/24'): ip_address('10.0.0.1'),
            ip_network('10.99.98.0/24'): ip_address('10.0.0.99')
        })
        self.assertEqual(vlan.route_count_by_ipv(4), 2)
        vlan.del_route(ip_network('10.99.99.0/24'))
        self.assertEqual(vlan.routes_by_ipv(4), {
            ip_network('10.99.98.0/24'): ip_address('10.0.0.99')
        })
        self.assertEqual(vlan.route_count_by_ipv(4), 1)
        vlan.del_route(ip_network('10.99.98.0/24'))
        self.assertEqual(vlan.route_count_by_ipv(4), 0)
        self.assertEqual(vlan.routes_by_ipv(4), {})

    def test_modify_routes_v6(self):
        """Tests the add_route() and remove_route() methods with IPv4 routes"""

        vlan = VLAN(1, 1, self.input_config)

        self.assertEqual(vlan.routes_by_ipv(6), {})
        vlan.add_route(ip_network('fc00::10:0/112'), ip_address('fc00::1:1'))
        vlan.add_route(ip_network('fc00::20:0/112'), ip_address('fc00::1:99'))
        self.assertEqual(vlan.routes_by_ipv(6), {
            ip_network('fc00::10:0/112'): ip_address('fc00::1:1'),
            ip_network('fc00::20:0/112'): ip_address('fc00::1:99')
        })
        self.assertEqual(vlan.route_count_by_ipv(6), 2)
        vlan.del_route(ip_network('fc00::10:0/112'))
        self.assertEqual(vlan.routes_by_ipv(6), {
            ip_network('fc00::20:0/112'): ip_address('fc00::1:99')
        })
        self.assertEqual(vlan.route_count_by_ipv(6), 1)
        vlan.del_route(ip_network('fc00::20:0/112'))
        self.assertEqual(vlan.route_count_by_ipv(6), 0)
        self.assertEqual(vlan.routes_by_ipv(6), {})

    def test_modify_routes_static_v4(self):
        """Tests the add_route() and remove_route() methods,
        starting with configured static routes for IPv4"""

        self.input_config.update({
            'routes': [
                {'route': {'ip_dst': '10.99.97.0/24', 'ip_gw': '10.0.0.99'}},
            ],
        })

        vlan = VLAN(1, 1, self.input_config)

        self.assertEqual(vlan.routes_by_ipv(4), {
            ip_network('10.99.97.0/24'): ip_address('10.0.0.99')
        })
        vlan.add_route(ip_network('10.99.99.0/24'), ip_address('10.0.0.1'))
        vlan.add_route(ip_network('10.99.98.0/24'), ip_address('10.0.0.99'))
        self.assertEqual(vlan.routes_by_ipv(4), {
            ip_network('10.99.99.0/24'): ip_address('10.0.0.1'),
            ip_network('10.99.98.0/24'): ip_address('10.0.0.99'),
            ip_network('10.99.97.0/24'): ip_address('10.0.0.99')
        })
        self.assertEqual(vlan.route_count_by_ipv(4), 3)
        vlan.del_route(ip_network('10.99.99.0/24'))
        self.assertEqual(vlan.routes_by_ipv(4), {
            ip_network('10.99.97.0/24'): ip_address('10.0.0.99'),
            ip_network('10.99.98.0/24'): ip_address('10.0.0.99')
        })
        self.assertEqual(vlan.route_count_by_ipv(4), 2)
        vlan.del_route(ip_network('10.99.98.0/24'))
        self.assertEqual(vlan.route_count_by_ipv(4), 1)
        self.assertEqual(vlan.routes_by_ipv(4), {
            ip_network('10.99.97.0/24'): ip_address('10.0.0.99')
        })

    def test_modify_routes_static_v6(self):
        """Tests the add_route() and remove_route() methods,
        starting with configured static routes for IPv6"""

        self.input_config.update({
            'routes': [
                {'route': {'ip_dst': 'fc00::30:0/112', 'ip_gw': 'fc00::1:99'}},
            ],
        })

        vlan = VLAN(1, 1, self.input_config)

        self.assertEqual(vlan.routes_by_ipv(6), {
            ip_network('fc00::30:0/112'): ip_address('fc00::1:99')
        })
        vlan.add_route(ip_network('fc00::10:0/112'), ip_address('fc00::1:1'))
        vlan.add_route(ip_network('fc00::20:0/112'), ip_address('fc00::1:99'))
        self.assertEqual(vlan.routes_by_ipv(6), {
            ip_network('fc00::10:0/112'): ip_address('fc00::1:1'),
            ip_network('fc00::20:0/112'): ip_address('fc00::1:99'),
            ip_network('fc00::30:0/112'): ip_address('fc00::1:99')
        })
        self.assertEqual(vlan.route_count_by_ipv(6), 3)
        vlan.del_route(ip_network('fc00::10:0/112'))
        self.assertEqual(vlan.routes_by_ipv(6), {
            ip_network('fc00::30:0/112'): ip_address('fc00::1:99'),
            ip_network('fc00::20:0/112'): ip_address('fc00::1:99')
        })
        self.assertEqual(vlan.route_count_by_ipv(6), 2)
        vlan.del_route(ip_network('fc00::20:0/112'))
        self.assertEqual(vlan.route_count_by_ipv(6), 1)
        self.assertEqual(vlan.routes_by_ipv(6), {
            ip_network('fc00::30:0/112'): ip_address('fc00::1:99')
        })



"""Configure routing between VLANs."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2018 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from faucet.conf import Conf, test_config_condition


class Router(Conf):
    """Implement FAUCET configuration for a router."""

    vlans = None

    defaults = {
        'vlans': None,
    }

    defaults_types = {
        'vlans': list,
    }

    def __str__(self):
        return self._id

    def check_config(self):
        super(Router, self).check_config()
        test_config_condition(not (isinstance(self.vlans, list) and len(self.vlans) > 1), (
            'router %s must have at least 2 VLANs configured' % self))



#!/usr/bin/env python3
import logging
import os
import sys
import Fake
from faucet import faucet
from ryu.controller import dpset
from faucet import faucet_experimental_api

logging.disable(logging.CRITICAL)

# run faucet
application = faucet.Faucet(dpset=dpset.DPSet(), faucet_experimental_api=faucet_experimental_api.FaucetExperimentalAPI())
application.start()

# make sure dps are running
for dp_id, valve in list(application.valves.items()):
    valve.dp.running = True

import afl
while afl.loop(1000):
    # receive input from afl
    rcv = sys.stdin.read()
    data = None
    try:
        data = bytearray.fromhex(rcv)
    except (ValueError, TypeError):
        pass

    if data is not None:
        # create fake packet
        dp = Fake.Datapath(1)
        msg = Fake.Message(datapath=dp, cookie=1524372928, port=1, data=data, in_port=1)
        pkt = Fake.RyuEvent(msg)

        # send fake packet to faucet
        application.packet_in_handler(pkt)

os._exit(0)


#!/usr/bin/env python3

import logging
import tempfile
import os
import sys
from faucet import config_parser as cp

LOGNAME = 'FAUCETLOG'

logging.disable(logging.CRITICAL)
tmpdir = tempfile.mkdtemp()

def create_config_file(config):
    conf_file_name = os.path.join(tmpdir, 'faucet.yaml')
    with open(conf_file_name, 'w') as conf_file:
        conf_file.write(config)
    return conf_file_name

import afl
while afl.loop(500):
    data = sys.stdin.read()
    file_name = create_config_file(data)
    try:
        cp.dp_parser(file_name, LOGNAME)
    except cp.InvalidConfigError as err:
        pass
os._exit(0)


"""Unit tests for FaucetBgp"""

import unittest
from unittest.mock import patch

from ipaddress import ip_address, ip_interface

from faucet.faucet_bgp import FaucetBgp

class FakeVLAN(object):
    """Fakes a VLAN object"""

    def __init__(self, vlan_config):
        self.dp_id = 1
        self.vid = 101
        self.vlan_config = vlan_config
        self.bgp_port = 1234
        self.bgp_as = 12345
        self.bgp_routerid = '1.1.1.1'
        self.bgp_neighbor_as = 54321

    def bgp_ipvs(self): # pylint: disable=missing-docstring
        return self.vlan_config['bgp_ipvs']

    def bgp_server_addresses_by_ipv(self, ipv): # pylint: disable=missing-docstring
        return self.vlan_config['bgp_server_addresses_by_ipv'][ipv]

    def faucet_vips_by_ipv(self, ipv): # pylint: disable=missing-docstring
        return self.vlan_config['faucet_vips_by_ipv'][ipv]

    def routes_by_ipv(self, ipv): # pylint: disable=missing-docstring
        return self.vlan_config['routes_by_ipv'][ipv]

    def bgp_neighbor_addresses_by_ipv(self, ipv): # pylint: disable=missing-docstring
        return self.vlan_config['bgp_neighbor_addresses_by_ipv'][ipv]

class FakeDP(object): # pylint: disable=too-few-public-methods
    """Fakes a DP object"""

    def __init__(self, vlans):
        self.vlans = vlans

    def bgp_vlans(self): # pylint: disable=missing-docstring
        return self.vlans

class FakeValve(object): # pylint: disable=too-few-public-methods
    """Fakes a Valve object"""

    def __init__(self, dp):
        self.dp = dp # pylint: disable=invalid-name

class FakeLogger(object): # pylint: disable=too-few-public-methods
    """Fakes a Ryu logger object"""

    def __init__(self):
        self.warning = None


class FaucetBgpTest(unittest.TestCase): # pytype: disable=module-attr
    """Test Faucet BGP"""

    def test_creates_one_ipv4_speaker(self):
        """Check that we create one IPv4 speaker"""

        vlan_config = {
            'bgp_server_addresses_by_ipv': {
                4: [ip_address('127.0.0.1')],
                6: []
            },
            'faucet_vips_by_ipv': {
                4: [ip_interface('10.0.0.1/24')],
                6: []
            },
            'routes_by_ipv': {
                4: {},
                6: {}
            },
            'bgp_neighbor_addresses_by_ipv': {
                4: [ip_address('172.0.0.2')],
                6: []
            },
            'bgp_ipvs': [4]
        }
        vlan = FakeVLAN(vlan_config)
        dp = FakeDP([vlan]) # pylint: disable=invalid-name
        valves = {'sw1': FakeValve(dp)}
        logger = FakeLogger()

        with patch('faucet.faucet_bgp.Beka') as beka:
            faucet_bgp = FaucetBgp(logger, None, None)
            faucet_bgp.reset(valves)

        self.assertEqual(beka.call_count, 1)

    def test_creates_one_ipv6_speaker(self):
        """Check that we create one IPv6 speaker"""

        vlan_config = {
            'bgp_server_addresses_by_ipv': {
                4: [],
                6: [ip_address('2001:db1::1')]
            },
            'faucet_vips_by_ipv': {
                4: [],
                6: [ip_interface('2001:db9::/32')]
            },
            'routes_by_ipv': {
                4: {},
                6: {}
            },
            'bgp_neighbor_addresses_by_ipv': {
                4: [],
                6: [ip_address('2001:db1::2')]
            },
            'bgp_ipvs': [6]
        }
        vlan = FakeVLAN(vlan_config)
        dp = FakeDP([vlan]) # pylint: disable=invalid-name
        valves = {'sw1': FakeValve(dp)}
        logger = FakeLogger()

        with patch('faucet.faucet_bgp.Beka') as beka:
            faucet_bgp = FaucetBgp(logger, None, None)
            faucet_bgp.reset(valves)

        self.assertEqual(beka.call_count, 1)

    def test_creates_dualstack_speakers(self):
        """Check that we create one IPv4 and one IPv6 speaker"""

        vlan_config = {
            'bgp_server_addresses_by_ipv': {
                4: [ip_address('127.0.0.1')],
                6: [ip_address('2001:db1::1')]
            },
            'faucet_vips_by_ipv': {
                4: [ip_interface('10.0.0.1/24')],
                6: [ip_interface('2001:db9::/32')]
            },
            'routes_by_ipv': {
                4: {},
                6: {}
            },
            'bgp_neighbor_addresses_by_ipv': {
                4: [ip_address('172.0.0.2')],
                6: [ip_address('2001:db1::2')]
            },
            'bgp_ipvs': [4, 6]
        }
        vlan = FakeVLAN(vlan_config)
        dp = FakeDP([vlan]) # pylint: disable=invalid-name
        valves = {'sw1': FakeValve(dp)}
        logger = FakeLogger()

        with patch('faucet.faucet_bgp.Beka') as beka:
            faucet_bgp = FaucetBgp(logger, None, None)
            faucet_bgp.reset(valves)

        self.assertEqual(beka.call_count, 2)



"""Configuration for ACLs."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2018 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from faucet.valve_of import MATCH_FIELDS, OLD_MATCH_FIELDS
from faucet.conf import Conf, test_config_condition


class ACL(Conf):
    """Contains the state for an ACL, including the configuration.

ACL Config

ACLs are configured under the 'acls' configuration block. The acls block
contains a dictionary of individual acls each keyed by its name.

Each acl contains a list of rules, a packet will have the first matching rule
applied to it.

Each rule is a dictionary containing the single key 'rule' with the value the
matches and actions for the rule.

The matches are key/values based on the ryu RESTFul API.
The key 'actions' contains a dictionary with keys/values as follows:

 * allow (int): if 1 allow the packet to continue through the Faucet \
       pipeline, if 0 drop the packet.
 * force_port_vlan (int): if 1, do not verify the VLAN/port association \
       for this packet and override any VLAN ACL on the forced VLAN.
 * meter (str): meter to apply to the packet
 * output (dict): used to output a packet directly. details below.
 * cookie (int): set flow cookie to this value on this flow

The output action contains a dictionary with the following elements:

 * port (int or string): the port to output the packet to
 * ports (list): a list of the ports (int or string) to output the packet to
 * set_fields (list): a list of fields to set with values
 * dl_dst (str): old style request to set eth_dst to a value (set_fields recommended)
 * pop_vlans: (int): pop the packet vlan before outputting
 * vlan_vid: (int): push the vlan vid on the packet when outputting
 * vlan_vids: (list): push the list of vlans on the packet when outputting, with option eth_type
 * swap_vid (int): rewrite the vlan vid of the packet when outputting
 * failover (dict): Output with a failover port (experimental)
"""

    rules = None
    exact_match = None
    defaults = {
        'rules': None,
        'exact_match': False,
    }
    defaults_types = {
        'rules': list,
        'exact_match': bool,
    }
    rule_types = {
        'cookie': int,
        'actions': dict,
        'description': str,
    }
    actions_types = {
        'meter': str,
        'mirror': (str, int),
        'output': dict,
        'allow': int,
        'force_port_vlan': int,
    }
    output_actions_types = {
        'port': (str, int),
        'ports': list,
        'failover': dict,
        'set_fields': list,
        'dl_dst': str,
        'pop_vlans': int,
        'swap_vid': int,
        'vlan_vid': int,
        'vlan_vids': list,
    }

    def __init__(self, _id, dp_id, conf):
        super(ACL, self).__init__(_id, dp_id, conf)
        rules = conf
        if isinstance(conf, dict):
            if 'exact_match' in conf and conf['exact_match']:
                self.exact_match = True
            test_config_condition('rules' not in conf, 'no rules found for ACL %s' % _id)
            rules = conf['rules']
        self.rules = []
        test_config_condition(not isinstance(rules, list), (
            "ACL rules is %s not %s" % (type(rules), dict)))
        for match_fields in (MATCH_FIELDS, OLD_MATCH_FIELDS):
            for match in list(match_fields.keys()):
                self.rule_types[match] = (str, int)
        for rule in rules:
            test_config_condition(not isinstance(rule, dict), (
                "ACL rule is %s not %s" % (type(rule), dict)))
            for rule_key, rule_content in list(rule.items()):
                test_config_condition(rule_key != 'rule', "Incorrect ACL rule key")
                test_config_condition(not isinstance(rule_content, dict), (
                    "ACL rule content is %s not %s") % (type(rule_content), dict))
                self._check_conf_types(rule_content, self.rule_types)
                for rule_field, rule_conf in list(rule_content.items()):
                    if rule_field == 'cookie':
                        test_config_condition(rule_conf < 0 or rule_conf > 2**16, (
                            'rule cookie value must be 0-2**16'))
                    elif rule_field == 'actions':
                        test_config_condition(not rule_conf, "Missing rule actions in ACL %s" % _id)
                        self._check_conf_types(rule_conf, self.actions_types)
                        for action_name, action_conf in list(rule_conf.items()):
                            if action_name == 'output':
                                self._check_conf_types(action_conf, self.output_actions_types)
                self.rules.append(rule_content)

    def to_conf(self):
        return [{'rule': rule} for rule in self.rules]
